# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Discovers, validates, and loads built-in and user extensions."""

from __future__ import annotations

import ast
import hashlib
import importlib.util
import os
import sys
import threading
import time
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from . import command_manager, debug, gsettings_registry
from .extension import Extension, get_translation

if TYPE_CHECKING:
    import gettext
    from collections.abc import Callable

_SCHEMA = "extensions"
_KEY_DISABLED = "disabled-extensions"
_KEY_APPROVED = "approved-user-extensions"
_KEY_SETTINGS = "settings"


class UserExtensionStatus(Enum):
    """Status of a discovered user extension file."""

    INVALID = "invalid"
    UNAPPROVED = "unapproved"
    APPROVED = "approved"
    MODIFIED = "modified"
    DISABLED = "disabled"


@dataclass(frozen=True)
class UserExtensionMetadata:
    """Metadata parsed from a user extension without executing it."""

    class_name: str | None = None
    group_label: str | None = None
    description: str = ""
    version: str = ""
    author: str = ""
    organization: str = ""
    copyright: str = ""
    website: str = ""


@dataclass(frozen=True)
class UserExtensionInfo:
    """Non-executing metadata for a user extension file."""

    filename: str
    filepath: str
    is_package: bool
    class_name: str | None
    group_label: str | None
    description: str
    version: str
    author: str
    organization: str
    copyright: str
    website: str
    file_hash: str
    approved_hash: str | None
    status: UserExtensionStatus

    @property
    def is_approved(self) -> bool:
        """Returns True if the current file hash is approved."""

        return self.approved_hash == self.file_hash


@gsettings_registry.get_registry().gsettings_schema("org.gnome.Orca.Extensions", name="extensions")
class ExtensionLoader:  # pylint: disable=too-many-public-methods
    """Discovers, validates, and loads built-in and user extensions."""

    _SHUTDOWN_HOOK_TOTAL_TIMEOUT_SECONDS = 0.5

    def __init__(self) -> None:
        self._builtins: list[tuple[Callable[[], Extension], str]] = []
        self._user_extensions: list[Extension] = []
        self._user_extensions_by_source: dict[str, Extension] = {}
        self._speech_output_handlers: list[Extension] = []
        self._braille_output_handlers: list[Extension] = []
        self._orca_is_ready = False

    @gsettings_registry.get_registry().gsetting(
        key="disabled-extensions",
        schema="extensions",
        gtype="as",
        default=[],
        summary="Extensions disabled by the user (by MODULE_NAME)",
    )
    def get_disabled_extensions(self) -> list[str]:
        """Returns the list of disabled extension MODULE_NAMEs."""

        return gsettings_registry.get_registry().layered_lookup(
            _SCHEMA,
            _KEY_DISABLED,
            "as",
            default=[],
        )

    def set_disabled_extensions(self, value: list[str]) -> bool:
        """Sets the list of disabled extension class names."""

        return gsettings_registry.get_registry().set_strv(
            _SCHEMA,
            _KEY_DISABLED,
            value,
        )

    @gsettings_registry.get_registry().gsetting(
        key="approved-user-extensions",
        schema="extensions",
        gtype="a{ss}",
        default={},
        summary="Approved user extensions (filename to SHA256 hash)",
    )
    def get_approved_extensions(self) -> dict[str, str]:
        """Returns the dict of approved extensions (filename -> sha256)."""

        return gsettings_registry.get_registry().layered_lookup(
            _SCHEMA,
            _KEY_APPROVED,
            "a{ss}",
            default={},
        )

    def set_approved_extensions(self, value: dict[str, str]) -> bool:
        """Sets the dict of approved extensions."""

        return gsettings_registry.get_registry().set_dict(
            _SCHEMA,
            _KEY_APPROVED,
            "a{ss}",
            value,
        )

    @gsettings_registry.get_registry().gsetting(
        key="settings",
        schema="extensions",
        gtype="a{sv}",
        default={},
        summary="Settings for a user extension",
    )
    def get_user_extension_settings(self) -> dict:
        """Returns settings for the extensions root path."""

        return gsettings_registry.get_registry().layered_lookup(
            _SCHEMA,
            _KEY_SETTINGS,
            "a{sv}",
            default={},
        )

    def approve_extension(self, filename: str, sha256_hash: str) -> None:
        """Approves an extension by recording its filename and hash."""

        approved = dict(self.get_approved_extensions())
        approved[filename] = sha256_hash
        self.set_approved_extensions(approved)

    def approve_extension_file(self, filepath: str) -> str:
        """Computes the hash and approves the extension. Returns the hash."""

        filename = self._get_source_id(filepath)
        file_hash = self._compute_hash(filepath)
        self.approve_extension(filename, file_hash)
        return file_hash

    def revoke_extension(self, filename: str) -> None:
        """Revokes approval for an extension."""

        approved = dict(self.get_approved_extensions())
        approved.pop(filename, None)
        self.set_approved_extensions(approved)

    def discover_user_extensions(self, extensions_dir: str) -> list[UserExtensionInfo]:
        """Returns metadata for user extension files without executing them."""

        if not os.path.isdir(extensions_dir):
            return []

        approved = self.get_approved_extensions()
        disabled = self.get_disabled_extensions()
        result: list[UserExtensionInfo] = []

        for filename, filepath, is_package in self._iter_extension_sources(extensions_dir):
            metadata = self.get_metadata(filepath)
            file_hash = self._compute_hash(filepath)
            approved_hash = approved.get(filename)
            status = self._get_user_extension_status(
                metadata.class_name,
                file_hash,
                approved_hash,
                disabled,
            )
            result.append(
                UserExtensionInfo(
                    filename=filename,
                    filepath=filepath,
                    is_package=is_package,
                    class_name=metadata.class_name,
                    group_label=metadata.group_label,
                    description=metadata.description,
                    version=metadata.version,
                    author=metadata.author,
                    organization=metadata.organization,
                    copyright=metadata.copyright,
                    website=metadata.website,
                    file_hash=file_hash,
                    approved_hash=approved_hash,
                    status=status,
                )
            )

        return result

    def reload_user_extension(self, extensions_dir: str, source_id: str) -> None:
        """Reloads one user extension while preserving unrelated instances."""

        old_extension = self._user_extensions_by_source.pop(source_id, None)
        if old_extension is not None:
            if self._orca_is_ready:
                self._run_lifecycle_hook(old_extension, old_extension.on_disabled)
            old_extension.disable()
            self._sync_user_extension_lists()

        self._remove_user_extension_modules(source_id)
        new_extension = self._load_user_extension_source(extensions_dir, source_id)
        if old_extension is None and new_extension is None:
            return

        # Re-register commands in source order so keybinding conflict winners
        # do not depend on which extension was reloaded.
        for extension in self._user_extensions_by_source.values():
            extension.reset_commands()

        if new_extension is not None:
            self._user_extensions_by_source[source_id] = new_extension
        self._sync_user_extension_lists()
        self.set_up_user_commands()
        command_manager.get_manager().activate_commands("reloaded user extension")
        if self._orca_is_ready and new_extension is not None:
            self._run_lifecycle_hook(new_extension, new_extension.on_enabled)

    def register_builtin(self, getter: Callable[[], Extension], group_label: str) -> None:
        """Registers a built-in extension with the loader."""

        self._builtins.append((getter, group_label))

    def discover_and_load(self, extensions_dir: str) -> None:
        """Scans the extensions directory and loads approved user extensions."""

        if not os.path.isdir(extensions_dir):
            msg = f"EXTENSION LOADER: Extensions directory not found: {extensions_dir}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        approved = self.get_approved_extensions()
        disabled = self.get_disabled_extensions()

        for filename, filepath, _is_package in self._iter_extension_sources(extensions_dir):
            extension = self._load_user_extension(filename, filepath, approved, disabled)
            if extension is None:
                continue

            self._user_extensions_by_source[filename] = extension

        self._sync_user_extension_lists()

    def _load_user_extension_source(
        self,
        extensions_dir: str,
        source_id: str,
    ) -> Extension | None:
        """Loads source_id if it is present, approved, and enabled."""

        if not os.path.isdir(extensions_dir):
            return None

        source = None
        for filename, filepath, _is_package in self._iter_extension_sources(extensions_dir):
            if filename == source_id:
                source = (filename, filepath)
                break
        if source is None:
            return None

        filename, filepath = source
        return self._load_user_extension(
            filename,
            filepath,
            self.get_approved_extensions(),
            self.get_disabled_extensions(),
        )

    def _load_user_extension(
        self,
        filename: str,
        filepath: str,
        approved: dict[str, str],
        disabled: list[str],
    ) -> Extension | None:
        """Loads an approved, enabled user-extension source."""

        file_hash = self._compute_hash(filepath)
        approved_hash = approved.get(filename)
        if approved_hash is None:
            msg = (
                f"EXTENSION LOADER: New extension found: {filename}. "
                f"Not approved. Hash: {file_hash}"
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            profile = gsettings_registry.get_registry().get_active_profile()
            msg = (
                f"EXTENSION LOADER: To approve, run: dconf write "
                f"/org/gnome/orca/{profile}/extensions/"
                f"approved-user-extensions "
                f"\"{{'{filename}': '{file_hash}'}}\""
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return None

        if approved_hash != file_hash:
            msg = (
                f"EXTENSION LOADER: Extension modified: {filename}. "
                f"Approved hash: {approved_hash} "
                f"Current hash: {file_hash}. "
                "Not loading until re-approved."
            )
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            profile = gsettings_registry.get_registry().get_active_profile()
            msg = (
                f"EXTENSION LOADER: To re-approve, run: dconf write "
                f"/org/gnome/orca/{profile}/extensions/"
                f"approved-user-extensions "
                f"\"{{'{filename}': '{file_hash}'}}\""
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return None

        if disabled:
            class_name = self.get_class_name(filepath)
            if class_name in disabled:
                msg = f"EXTENSION LOADER: Extension {class_name} is disabled. Skipping."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return None

        extension = self._load_from_file(filepath, filename)
        if extension is None:
            return None

        # Static metadata can differ from the subclass selected when the module is loaded.
        if extension.module_name in disabled:
            msg = f"EXTENSION LOADER: Extension {extension.module_name} is disabled. Skipping."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            extension.disable()
            self._remove_user_extension_modules(filename)
            return None

        msg = f"EXTENSION LOADER: Loaded extension {extension.module_name} from {filename}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return extension

    def _sync_user_extension_lists(self) -> None:
        """Rebuilds the user-extension and output-handler lists in source order."""

        self._user_extensions = [
            self._user_extensions_by_source[source_id]
            for source_id in sorted(self._user_extensions_by_source)
        ]
        self._speech_output_handlers = [
            extension
            for extension in self._user_extensions
            if self._extension_handles_speech_output(extension)
        ]
        self._braille_output_handlers = [
            extension
            for extension in self._user_extensions
            if self._extension_handles_braille_output(extension)
        ]

    def set_up_all_commands(self) -> None:
        """Calls set_up_commands on all enabled extensions (built-in and user)."""

        disabled = self.get_disabled_extensions()

        for getter, _group_label in self._builtins:
            ext = getter()
            if ext.module_name in disabled:
                ext.disable()
                continue
            msg = f"EXTENSION LOADER: Loading built-in extension {ext.module_name}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            ext.set_up_commands()

        self.set_up_user_commands()

    def set_up_user_commands(self) -> None:
        """Calls set_up_commands on all enabled user extensions."""

        disabled = self.get_disabled_extensions()

        for ext in self._user_extensions:
            if ext.module_name in disabled:
                ext.disable()
                continue
            msg = f"EXTENSION LOADER: Loading user extension {ext.module_name}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            ext.set_up_commands()

    def notify_user_extensions_ready(self) -> None:
        """Notifies loaded user extensions that Orca is ready."""

        if self._orca_is_ready:
            return
        self._orca_is_ready = True
        for extension in self._user_extensions:
            self._run_lifecycle_hook(extension, extension.on_ready)

    @staticmethod
    def _run_lifecycle_hook(
        extension: Extension,
        hook: Callable[[], None],
    ) -> None:
        """Runs a lifecycle hook for one user extension, logging exceptions."""

        msg = f"EXTENSION LOADER: Calling {extension.module_name}.{hook.__name__}()"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        try:
            hook()
        except Exception:  # pylint: disable=broad-exception-caught
            debug.print_exception(debug.LEVEL_WARNING)

    def shutdown_user_extensions(self) -> None:
        """Notifies loaded user extensions that Orca is shutting down."""

        threads: list[tuple[Extension, threading.Thread]] = []
        for ext in self._user_extensions:
            msg = f"EXTENSION LOADER: Shutting down user extension {ext.module_name}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            thread = threading.Thread(target=self._run_shutdown_hook, args=(ext,), daemon=True)
            thread.start()
            threads.append((ext, thread))

        deadline = time.monotonic() + self._SHUTDOWN_HOOK_TOTAL_TIMEOUT_SECONDS
        for _ext, thread in threads:
            timeout = deadline - time.monotonic()
            if timeout <= 0:
                break
            thread.join(timeout)

        timed_out = [ext.module_name for ext, thread in threads if thread.is_alive()]
        if timed_out:
            names = ", ".join(timed_out)
            msg = f"EXTENSION LOADER: Shutdown hook timed out for: {names}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)

    @staticmethod
    def _run_shutdown_hook(extension: Extension) -> None:
        """Runs an extension shutdown hook, logging exceptions."""

        try:
            extension.on_shutdown()
        except Exception:  # pylint: disable=broad-exception-caught
            debug.print_exception(debug.LEVEL_WARNING)

    def iter_speech_output_handlers(self) -> list[Extension]:
        """Returns enabled user extensions that can handle outgoing speech."""

        if not self._speech_output_handlers:
            return []

        disabled = self.get_disabled_extensions()

        return [ext for ext in self._speech_output_handlers if ext.module_name not in disabled]

    def iter_braille_output_handlers(self) -> list[Extension]:
        """Returns enabled user extensions that can handle outgoing braille."""

        if not self._braille_output_handlers:
            return []

        disabled = self.get_disabled_extensions()

        return [ext for ext in self._braille_output_handlers if ext.module_name not in disabled]

    def get_loaded_user_extension(self, class_name: str) -> Extension | None:
        """Returns the loaded user extension with class_name."""

        return next(
            (ext for ext in self._user_extensions if ext.module_name == class_name),
            None,
        )

    @staticmethod
    def _extension_handles_speech_output(extension: Extension) -> bool:
        """Returns True if extension overrides the speech-output hook."""

        return type(extension).on_speech_output is not Extension.on_speech_output

    @staticmethod
    def _extension_handles_braille_output(extension: Extension) -> bool:
        """Returns True if extension overrides the braille-output hook."""

        return type(extension).on_braille_output is not Extension.on_braille_output

    @staticmethod
    def _get_user_extension_status(
        class_name: str | None,
        file_hash: str,
        approved_hash: str | None,
        disabled: list[str],
    ) -> UserExtensionStatus:
        """Returns the status for a discovered user extension file."""

        if class_name is None:
            return UserExtensionStatus.INVALID
        if approved_hash is None:
            return UserExtensionStatus.UNAPPROVED
        if approved_hash != file_hash:
            return UserExtensionStatus.MODIFIED
        if class_name in disabled:
            return UserExtensionStatus.DISABLED
        return UserExtensionStatus.APPROVED

    @staticmethod
    def get_class_name(filepath: str) -> str | None:
        """Returns the Extension subclass name from a file without executing it."""

        return ExtensionLoader.get_metadata(filepath).class_name

    @staticmethod
    def get_metadata(filepath: str) -> UserExtensionMetadata:
        """Returns metadata from an Extension subclass without executing it."""

        result = UserExtensionMetadata()
        source_path = ExtensionLoader._get_source_file(filepath)
        if source_path is None:
            return result

        try:
            with open(source_path, encoding="utf-8") as f:
                tree = ast.parse(f.read())
        except (OSError, SyntaxError):
            return result

        translation = get_translation(source_path)
        for node in ast.walk(tree):
            if not isinstance(node, ast.ClassDef):
                continue
            for base in node.bases:
                if isinstance(base, ast.Attribute):
                    name = base.attr
                elif isinstance(base, ast.Name):
                    name = base.id
                else:
                    continue
                if name == "Extension":
                    return UserExtensionMetadata(
                        class_name=node.name,
                        group_label=ExtensionLoader._get_class_string_constant(
                            node, "GROUP_LABEL", translation
                        ),
                        description=ExtensionLoader._get_class_string_constant(
                            node, "DESCRIPTION", translation
                        )
                        or "",
                        version=ExtensionLoader._get_class_string_constant(node, "VERSION") or "",
                        author=ExtensionLoader._get_class_string_constant(node, "AUTHOR") or "",
                        organization=ExtensionLoader._get_class_string_constant(
                            node, "ORGANIZATION"
                        )
                        or "",
                        copyright=ExtensionLoader._get_class_string_constant(node, "COPYRIGHT")
                        or "",
                        website=ExtensionLoader._get_class_string_constant(node, "WEBSITE") or "",
                    )

        return result

    @staticmethod
    def _get_class_string_constant(
        node: ast.ClassDef,
        name: str,
        translation: gettext.NullTranslations | None = None,
    ) -> str | None:
        """Returns a possibly localized string assigned in a class body."""

        for child in node.body:
            value: ast.expr | None = None
            if isinstance(child, ast.Assign):
                if not any(
                    isinstance(target, ast.Name) and target.id == name for target in child.targets
                ):
                    continue
                value = child.value
            elif isinstance(child, ast.AnnAssign):
                if not isinstance(child.target, ast.Name) or child.target.id != name:
                    continue
                value = child.value

            message, context = ExtensionLoader._get_translatable_literal(value)
            if message is None:
                continue
            if translation is None:
                return message
            if context is not None:
                return translation.pgettext(context, message)
            return translation.gettext(message)

        return None

    @staticmethod
    def _get_translatable_literal(value: ast.expr | None) -> tuple[str | None, str | None]:
        """Returns the literal message and optional context from an expression."""

        if isinstance(value, ast.Constant) and isinstance(value.value, str):
            return value.value, None
        if not isinstance(value, ast.Call):
            return None, None

        function_name = getattr(value.func, "id", getattr(value.func, "attr", None))
        if function_name in ("_", "gettext") and len(value.args) == 1:
            message = value.args[0]
            if isinstance(message, ast.Constant) and isinstance(message.value, str):
                return message.value, None

        if function_name == "pgettext" and len(value.args) == 2:
            context, message = value.args
            if (
                isinstance(context, ast.Constant)
                and isinstance(context.value, str)
                and isinstance(message, ast.Constant)
                and isinstance(message.value, str)
            ):
                return message.value, context.value

        return None, None

    @staticmethod
    def _compute_hash(filepath: str) -> str:
        """Returns the SHA256 hex digest of an extension file or package."""

        if os.path.isdir(filepath):
            return ExtensionLoader._compute_package_hash(filepath)

        sha256 = hashlib.sha256()
        with open(filepath, "rb") as f:
            for chunk in iter(lambda: f.read(8192), b""):
                sha256.update(chunk)
        return sha256.hexdigest()

    @staticmethod
    def _compute_package_hash(dirpath: str) -> str:
        """Returns a deterministic SHA256 digest for a package extension."""

        sha256 = hashlib.sha256()
        for root, dirs, files in os.walk(dirpath):
            dirs[:] = sorted(
                dirname
                for dirname in dirs
                if dirname != "__pycache__" and not dirname.startswith(".")
            )
            for basename in sorted(files):
                if ExtensionLoader._should_ignore_package_file(basename):
                    continue
                path = os.path.join(root, basename)
                relpath = os.path.relpath(path, dirpath).replace(os.sep, "/")
                sha256.update(relpath.encode("utf-8"))
                sha256.update(b"\0")
                with open(path, "rb") as f:
                    for chunk in iter(lambda: f.read(8192), b""):
                        sha256.update(chunk)
                sha256.update(b"\0")
        return sha256.hexdigest()

    @staticmethod
    def _should_ignore_package_file(filename: str) -> bool:
        """Returns True if filename should not affect package approval."""

        return filename.startswith(".") or filename.endswith((".pyc", ".pyo", "~"))

    @staticmethod
    def _iter_extension_sources(extensions_dir: str) -> list[tuple[str, str, bool]]:
        """Returns extension source ids, paths, and package flags."""

        sources: list[tuple[str, str, bool]] = []
        for filename in sorted(os.listdir(extensions_dir)):
            if filename.startswith("_"):
                continue

            filepath = os.path.join(extensions_dir, filename)
            if os.path.isfile(filepath) and filename.endswith(".py"):
                sources.append((filename, filepath, False))
            elif os.path.isdir(filepath):
                init_path = os.path.join(filepath, "__init__.py")
                if os.path.isfile(init_path):
                    sources.append((filename, filepath, True))
        return sources

    @staticmethod
    def _get_source_id(filepath: str) -> str:
        """Returns the approval id for an extension source."""

        return os.path.basename(filepath)

    @staticmethod
    def _get_user_extension_namespace(source_id: str) -> str:
        """Returns the Python namespace for a user-extension source id."""

        if source_id.endswith(".py"):
            return os.path.splitext(source_id)[0]
        return source_id

    @staticmethod
    def _remove_user_extension_modules(source_id: str) -> None:
        """Removes a user extension and its submodules from the module cache."""

        namespace = ExtensionLoader._get_user_extension_namespace(source_id)
        module_name = f"orca_user_extension.{namespace}"
        names = [
            name
            for name in sys.modules
            if name == module_name or name.startswith(f"{module_name}.")
        ]
        for name in names:
            sys.modules.pop(name, None)

    @staticmethod
    def _get_source_file(filepath: str) -> str | None:
        """Returns the Python file that defines extension metadata."""

        if os.path.isdir(filepath):
            init_path = os.path.join(filepath, "__init__.py")
            if os.path.isfile(init_path):
                return init_path
            return None
        if os.path.isfile(filepath):
            return filepath
        return None

    @staticmethod
    def _load_from_file(filepath: str, filename: str) -> Extension | None:
        """Loads an Extension subclass from a Python file or package."""

        source_file = ExtensionLoader._get_source_file(filepath)
        if source_file is None:
            msg = f"EXTENSION LOADER: No loadable source found for {filepath}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return None

        namespace = ExtensionLoader._get_user_extension_namespace(filename)
        module_name = f"orca_user_extension.{namespace}"
        try:
            if os.path.isdir(filepath):
                spec = importlib.util.spec_from_file_location(
                    module_name,
                    source_file,
                    submodule_search_locations=[filepath],
                )
            else:
                spec = importlib.util.spec_from_file_location(module_name, source_file)
            if spec is None or spec.loader is None:
                msg = f"EXTENSION LOADER: Could not create spec for {filepath}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)
                return None

            module = importlib.util.module_from_spec(spec)
            # The module must be in sys.modules before exec_module so that code run during
            # class creation (e.g. @dataclass) can look itself up via sys.modules[__name__].
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
        except Exception as error:  # pylint: disable=broad-exception-caught
            ExtensionLoader._remove_user_extension_modules(filename)
            msg = f"EXTENSION LOADER: Failed to load {filepath}: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return None

        for attr_name in dir(module):
            obj = module.__dict__.get(attr_name)
            if isinstance(obj, type) and issubclass(obj, Extension) and obj is not Extension:
                try:
                    instance = obj()
                    instance.mark_as_user_extension(namespace)
                    return instance
                except Exception as error:  # pylint: disable=broad-exception-caught
                    msg = (
                        f"EXTENSION LOADER: Failed to instantiate "
                        f"{attr_name} from {filepath}: {error}"
                    )
                    debug.print_message(debug.LEVEL_WARNING, msg, True)
                    return None

        msg = f"EXTENSION LOADER: No Extension subclass found in {filepath}"
        debug.print_message(debug.LEVEL_WARNING, msg, True)
        return None


_loader: ExtensionLoader = ExtensionLoader()


def get_loader() -> ExtensionLoader:
    """Returns the ExtensionLoader singleton."""

    return _loader



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Utilities for accessible documents."""

from __future__ import annotations

import urllib.parse

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import messages
from .ax_collection import AXCollection
from .ax_document import AXDocument
from .ax_utilities_role import AXUtilitiesRole
from .ax_utilities_state import AXUtilitiesState
from .ax_utilities_table import AXUtilitiesTable


class AXUtilitiesDocument:
    """Utilities for accessible documents."""

    @staticmethod
    def get_uri(document: Atspi.Accessible) -> str:
        """Returns the uri of document."""

        attributes = AXDocument.get_attributes_dict(document)
        return attributes.get("DocURL", attributes.get("URI", ""))

    @staticmethod
    def get_mime_type(document: Atspi.Accessible) -> str:
        """Returns the mime type of document."""

        return AXDocument.get_attributes_dict(document).get("MimeType", "")

    @staticmethod
    def get_document_uri_fragment(document: Atspi.Accessible) -> str:
        """Returns the fragment portion of document's uri."""

        return urllib.parse.urlparse(AXUtilitiesDocument.get_uri(document)).fragment

    @staticmethod
    def is_plain_text(document: Atspi.Accessible) -> bool:
        """Returns True if document is a plain-text document."""

        mime_type = AXUtilitiesDocument.get_mime_type(document)
        if mime_type == "text/plain":
            return True
        if mime_type == "text/html":
            return AXUtilitiesDocument.get_uri(document).endswith(".txt")
        return False

    @staticmethod
    def is_pdf(document: Atspi.Accessible) -> bool:
        """Returns True if document is a PDF document."""

        mime_type = AXUtilitiesDocument.get_mime_type(document)
        if mime_type == "application/pdf":
            return True
        if mime_type == "text/html":
            return AXUtilitiesDocument.get_uri(document).endswith(".pdf")
        return False

    @staticmethod
    def _get_object_counts(document: Atspi.Accessible) -> dict[str, int]:
        """Returns a dictionary of object counts used in a document summary."""

        result = {
            "forms": 0,
            "landmarks": 0,
            "headings": 0,
            "tables": 0,
            "unvisited_links": 0,
            "visited_links": 0,
        }

        roles = [
            Atspi.Role.HEADING,
            Atspi.Role.LINK,
            Atspi.Role.TABLE,
            Atspi.Role.FORM,
            Atspi.Role.LANDMARK,
        ]

        rule = AXCollection.create_match_rule(
            roles=roles,
            role_match_type=Atspi.CollectionMatchType.ANY,
        )
        matches = AXCollection.get_all_matches(document, rule)

        for obj in matches:
            if AXUtilitiesRole.is_heading(obj):
                result["headings"] += 1
            elif AXUtilitiesRole.is_form(obj):
                result["forms"] += 1
            elif AXUtilitiesRole.is_table(obj) and not AXUtilitiesTable.is_layout_table(obj):
                result["tables"] += 1
            elif AXUtilitiesRole.is_link(obj):
                if AXUtilitiesState.is_visited(obj):
                    result["visited_links"] += 1
                else:
                    result["unvisited_links"] += 1
            elif AXUtilitiesRole.is_landmark(obj):
                result["landmarks"] += 1

        return result

    @staticmethod
    def get_document_summary(document: Atspi.Accessible, only_if_found: bool = True) -> str:
        """Returns a string summarizing the document's structure and objects of interest."""

        result = []
        counts = AXUtilitiesDocument._get_object_counts(document)
        result.append(messages.landmark_count(counts.get("landmarks", 0), only_if_found))
        result.append(messages.heading_count(counts.get("headings", 0), only_if_found))
        result.append(messages.form_count(counts.get("forms", 0), only_if_found))
        result.append(messages.table_count(counts.get("tables", 0), only_if_found))
        result.append(messages.visited_link_count(counts.get("visited_links", 0), only_if_found))
        result.append(
            messages.unvisited_link_count(counts.get("unvisited_links", 0), only_if_found),
        )
        result = list(filter(lambda x: x, result))
        if not result:
            return ""

        return messages.PAGE_SUMMARY_PREFIX % ", ".join(result)



# Orca
#
# Copyright 2005-2008 Google Inc.
# Portions Copyright 2007-2008, Sun Microsystems, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.
#
"""ACSS --- Aural CSS.

Class ACSS defines a simple wrapper for holding ACSS voice
definitions.  Speech engines implement the code for converting
ACSS definitions into engine-specific markup codes.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import ClassVar


class ACSS(dict):
    """Holds ACSS representation of a voice."""

    FAMILY = "family"
    RATE = "rate"
    GAIN = "gain"
    AVERAGE_PITCH = "average-pitch"
    PITCH_RANGE = "pitch-range"
    STRESS = "stress"
    RICHNESS = "richness"
    PUNCTUATIONS = "punctuations"
    VOICE_TYPE = "voice-type"

    settings: ClassVar[dict[str, str | int | None]] = {
        FAMILY: None,  # None means use the engine's default value.
        RATE: 50,
        GAIN: 10,
        AVERAGE_PITCH: 5,
        PITCH_RANGE: 5,
        STRESS: 5,
        RICHNESS: 5,
        PUNCTUATIONS: "all",
    }

    def __init__(self, props: dict | None = None):
        """Create and initialize ACSS structure."""

        super().__init__()
        if not props:
            self["established"] = False
            return
        for key, value in props.items():
            if key == "established" or key in ACSS.settings:
                self[key] = dict(value) if key == ACSS.FAMILY else value

    def __eq__(self, other):
        if not isinstance(other, ACSS):
            return False
        if self.get(ACSS.FAMILY) != other.get(ACSS.FAMILY):
            return False
        if self.get(ACSS.RATE) != other.get(ACSS.RATE):
            return False
        return self.get(ACSS.AVERAGE_PITCH) == other.get(ACSS.AVERAGE_PITCH)

    def __ne__(self, other):
        return not self.__eq__(other)

    # Mutable: dict items are modified after construction.
    __hash__ = dict.__hash__

    def update(self, new_dict):
        family = new_dict.get(ACSS.FAMILY)
        if isinstance(family, dict) and family.get("name") is None:
            new_dict.pop(ACSS.FAMILY)

        return super().update(new_dict)



# Orca
#
# Copyright (C) 2015 Igalia, S.L.
#
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Produces speech presentation for accessible objects."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from orca import debug, speech_generator
from orca.ax_object import AXObject
from orca.ax_utilities import AXUtilities
from orca.scripts import web

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from . import script


class SpeechGenerator(web.SpeechGenerator, speech_generator.SpeechGenerator):
    """Produces speech presentation for accessible objects."""

    # Type annotation to override the base class script type
    _script: script.Script

    @staticmethod
    def log_generator_output(func):
        """Decorator for logging."""

        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            tokens = [f"EVOLUTION SPEECH GENERATOR: {func.__name__}:", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return result

        return wrapper

    @log_generator_output
    def _generate_state_checked_for_cell(self, obj: Atspi.Accessible) -> list[Any]:
        if self._script.utilities.is_message_list_status_cell(obj):
            return []

        if self._script.utilities.is_message_list_toggle_cell(obj):
            if AXUtilities.cell_row_changed(obj) or not AXUtilities.is_focused(obj):
                return []

        return super()._generate_state_checked_for_cell(obj)

    @log_generator_output
    def _generate_accessible_label(self, obj: Atspi.Accessible) -> list[Any]:
        if self._script.utilities.is_message_list_toggle_cell(obj):
            return []

        return super()._generate_accessible_label(obj)

    @log_generator_output
    def _generate_accessible_name(self, obj: Atspi.Accessible) -> list[Any]:
        if self._script.utilities.is_message_list_toggle_cell(
            obj,
        ) and not self._script.utilities.is_message_list_status_cell(obj):
            return []

        return super()._generate_accessible_name(obj)

    @log_generator_output
    def _generate_real_active_descendant_displayed_text(
        self,
        obj: Atspi.Accessible,
    ) -> list[Any]:
        if self._script.utilities.is_message_list_toggle_cell(
            obj,
        ) and not self._script.utilities.is_message_list_status_cell(obj):
            if not AXUtilities.is_checked(obj):
                return []
            if AXUtilities.is_focused(obj) and not AXUtilities.cell_row_changed(obj):
                return []

        return super()._generate_real_active_descendant_displayed_text(obj)

    @log_generator_output
    def _generate_accessible_role(self, obj: Atspi.Accessible) -> list[Any]:
        if self._script.utilities.is_message_list_toggle_cell(obj) and not AXUtilities.is_focused(
            obj,
        ):
            return []

        return super()._generate_accessible_role(obj)

    @log_generator_output
    def _generate_state_unselected(self, obj: Atspi.Accessible) -> list[Any]:
        if self._script.utilities.is_message_list_toggle_cell(obj) or AXUtilities.is_tree_table(
            AXObject.get_parent(obj),
        ):
            return []

        return super()._generate_state_unselected(obj)



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=no-member
# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
# pylint: disable=too-many-lines

"""Displays a GUI for the user to set Orca preferences."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import (
    Gdk,  # pylint: disable=no-name-in-module
    GLib,
    GObject,
    Gtk,  # pylint: disable=no-name-in-module
)

from . import (
    braille_presenter,
    chat_presenter,
    command_manager,
    debug,
    document_presenter,
    event_manager,
    extension_loader_preferences_grid,
    flat_review_presenter,
    focus_manager,
    gsettings_registry,
    guilabels,
    learn_mode_presenter,
    messages,
    mouse_presenter,
    orca,
    orca_gui_helpers,
    preferences_grid_base,
    presentation_manager,
    profile_manager,
    pronunciation_dictionary_manager,
    say_all_presenter,
    sleep_mode_manager,
    sound_presenter,
    speech_presenter,
    spellcheck_presenter,
    system_information_presenter,
    text_attribute_manager,
    typing_echo_presenter,
)
from .ax_object import AXObject

if TYPE_CHECKING:
    from .scripts import default


def show_app_preferences_gui(current_script: default.Script) -> bool:
    """Shows the application-specific preferences window."""

    ui = OrcaSetupGUI(current_script)
    ui.show_gui()
    return True


def show_preferences_gui(default_script: default.Script) -> bool:
    """Shows the global preferences window."""

    ui = OrcaSetupGUI(default_script)
    ui.show_gui()
    return True


def close_preferences_gui_for_shutdown() -> None:
    """Closes the preferences window and its transient dialogs during shutdown."""

    if OrcaSetupGUI.WINDOW is not None:
        OrcaSetupGUI.WINDOW.close_for_shutdown()


class NavigationRow(Gtk.ListBoxRow):
    """ListBoxRow with a panel_id attribute for navigation."""

    def __init__(self, panel_id: str | None = None) -> None:
        super().__init__()
        self.panel_id = panel_id


class OrcaSetupGUI(Gtk.ApplicationWindow):  # pylint: disable=too-many-instance-attributes
    """Preferences window for configuring Orca screen reader settings."""

    WINDOW: OrcaSetupGUI | None = None

    _EVENTS_TO_SUSPEND: tuple[str, ...] = (
        "object:state-changed:checked",
        "object:state-changed:sensitive",
        "object:state-changed:showing",
        "object:children-changed:add",
        "object:children-changed:remove",
        "object:selection-changed",
        "object:property-change:accessible-name",
        "object:property-change:accessible-description",
    )

    def __init__(self, script: default.Script) -> None:
        if OrcaSetupGUI.WINDOW is not None:
            return

        msg = "PREFERENCES: Initializing UI"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        appearance_refs = orca_gui_helpers.sync_appearance()
        super().__init__(title=guilabels.DIALOG_SCREEN_READER_PREFERENCES)
        self._appearance_refs = appearance_refs

        self.connect("destroy", self.window_destroyed)
        self.connect("delete-event", self.window_closed)

        self._profile_name: str = profile_manager.get_manager().get_active_profile()
        self.script = script
        self._app_name: str | None = None
        if script.app is not None:
            self._app_name = AXObject.get_name(script.app) or None
        self._current_page_title: str = ""
        self._current_panel_id: str | None = None
        self._original_profile: str = profile_manager.get_manager().get_active_profile()

        titlebar_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)

        self.left_headerbar = Gtk.HeaderBar()
        self.left_headerbar.set_show_close_button(True)
        self.left_headerbar.set_title(guilabels.DIALOG_SCREEN_READER_PREFERENCES)
        self.left_headerbar.get_style_context().add_class("orca-left-headerbar")

        self._app_icon = Gtk.Image.new_from_icon_name("orca", Gtk.IconSize.MENU)
        self._app_icon.set_margin_start(2)

        self.menu_button = Gtk.MenuButton()
        menu_image = Gtk.Image.new_from_icon_name("open-menu-symbolic", Gtk.IconSize.BUTTON)
        self.menu_button.set_image(menu_image)
        self.menu_button.get_accessible().set_name(guilabels.MENU_BUTTON_OPTIONS)

        popover = Gtk.Popover()
        menu_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)

        help_item = Gtk.ModelButton()
        help_item.set_property("text", guilabels.DIALOG_HELP)
        help_item.connect("clicked", lambda _: self.help_button_clicked(None))
        menu_box.pack_start(help_item, False, False, 0)

        self.apply_item = Gtk.ModelButton()
        self.apply_item.set_property("text", guilabels.DIALOG_APPLY)
        self.apply_item.connect("clicked", lambda _: self.apply_button_clicked(None))
        menu_box.pack_start(self.apply_item, False, False, 0)

        self.save_item = Gtk.ModelButton()
        self.save_item.set_property("text", guilabels.BTN_SAVE)
        self.save_item.connect("clicked", lambda _: self.ok_button_clicked(None))
        menu_box.pack_start(self.save_item, False, False, 0)

        # Save Profile As is only for global preferences (profiles are global)
        if not self._app_name:
            save_as_item = Gtk.ModelButton()
            save_as_item.set_property("text", guilabels.PROFILE_SAVE_AS_TITLE)
            save_as_item.connect("clicked", lambda _: self._on_save_profile_as())
            menu_box.pack_start(save_as_item, False, False, 0)

        cancel_item = Gtk.ModelButton()
        cancel_item.set_property("text", guilabels.DIALOG_CANCEL)
        cancel_item.connect("clicked", lambda _: self.cancel_button_clicked(None))
        menu_box.pack_start(cancel_item, False, False, 0)

        menu_box.show_all()
        popover.add(menu_box)
        self.menu_button.set_popover(popover)

        self.left_headerbar.pack_end(self.menu_button)

        titlebar_box.pack_start(self.left_headerbar, False, False, 0)

        titlebar_separator = Gtk.Separator(orientation=Gtk.Orientation.VERTICAL)
        titlebar_box.pack_start(titlebar_separator, False, False, 0)

        self.panel_headerbar = Gtk.HeaderBar()
        self.panel_headerbar.set_show_close_button(True)
        self.panel_headerbar.set_title("")
        self.panel_headerbar.get_style_context().add_class("orca-panel-headerbar")

        titlebar_box.pack_start(self.panel_headerbar, True, True, 0)

        self._apply_decoration_layout()
        gtk_settings = Gtk.Settings.get_default()  # pylint: disable=no-value-for-parameter
        if gtk_settings is not None:
            gtk_settings.connect(
                "notify::gtk-decoration-layout",
                lambda *_: self._apply_decoration_layout(),
            )

        self.set_titlebar(titlebar_box)

        main_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)

        self.hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)

        self.sidebar_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self.sidebar_vbox.get_style_context().add_class("orca-sidebar")

        self.sidebar_scrolled = Gtk.ScrolledWindow()
        self.sidebar_scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.listbox = Gtk.ListBox()
        self.listbox.set_selection_mode(Gtk.SelectionMode.SINGLE)
        self.listbox.get_accessible().set_name(guilabels.PREFERENCES_CATEGORIES)
        self.listbox.connect("row-selected", self._on_row_selected)
        self.sidebar_scrolled.add(self.listbox)
        self.sidebar_vbox.pack_start(self.sidebar_scrolled, True, True, 0)

        self.listbox.connect("key-press-event", self._on_listbox_key_press)

        self.hbox.pack_start(self.sidebar_vbox, False, False, 0)

        separator = Gtk.Separator(orientation=Gtk.Orientation.VERTICAL)
        self.hbox.pack_start(separator, False, False, 0)

        self.content_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)

        self.stack = Gtk.Stack()
        self.stack.set_hexpand(True)
        self.stack.set_vhomogeneous(False)
        stack_scrolled = Gtk.ScrolledWindow()
        stack_scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        stack_scrolled.add(self.stack)
        self.content_vbox.pack_start(stack_scrolled, True, True, 0)

        self.hbox.pack_start(self.content_vbox, True, True, 0)

        main_vbox.pack_start(self.hbox, True, True, 0)

        self.add(main_vbox)

        # Ignore runtime overrides while creating preference grids so that
        # on-the-fly changes (e.g. speech rate) don't influence the values shown.
        registry = gsettings_registry.get_registry()
        registry.set_ignore_runtime(True)

        # Create all preference grids
        # Profiles grid controls are insensitive for app-specific prefs
        prof_manager = profile_manager.get_manager()
        self.profiles_grid = prof_manager.create_preferences_grid(
            profile_loaded_callback=self._on_profile_loaded,
            is_app_specific=bool(self._app_name),
            labels_update_callback=self.update_menu_labels,
            unsaved_changes_checker=self._has_unsaved_changes,
        )
        self.stack.add_named(self.profiles_grid, "profiles")
        self._add_navigation_row("profiles", self.profiles_grid.get_title())
        self.update_menu_labels()

        if not self._app_name:
            sleep_mgr = sleep_mode_manager.get_manager()
            self.sleep_mode_grid = sleep_mgr.create_preferences_grid()
            self.stack.add_named(self.sleep_mode_grid, "sleep_mode")
            self._add_navigation_row("sleep_mode", self.sleep_mode_grid.get_title())

        speech_pres = speech_presenter.get_presenter()

        def update_title(title: str) -> None:
            """Update the panel header title and window accessible name."""

            self._set_page_title(title)

        self.speech_grid = speech_pres.create_speech_preferences_grid(
            title_change_callback=update_title,
            app_name=self._app_name or "",
        )
        self.stack.add_named(self.speech_grid, "speech")
        self._add_navigation_row("speech", self.speech_grid.get_title())

        braille_pres = braille_presenter.get_presenter()
        self.braille_grid = braille_pres.create_preferences_grid(title_change_callback=update_title)
        self.stack.add_named(self.braille_grid, "braille")
        self._add_navigation_row("braille", self.braille_grid.get_title())

        sound_pres = sound_presenter.get_presenter()
        self.sound_grid = sound_pres.create_preferences_grid(title_change_callback=update_title)
        self.stack.add_named(self.sound_grid, "sound")
        self._add_navigation_row("sound", self.sound_grid.get_title())

        cmd_manager = command_manager.get_manager()
        self.keybindings_grid = cmd_manager.create_preferences_grid(
            self.script,
            title_change_callback=update_title,
        )
        self.stack.add_named(self.keybindings_grid, "keybindings")
        self._add_navigation_row("keybindings", self.keybindings_grid.get_title())

        typing_pres = typing_echo_presenter.get_presenter()
        self.typing_echo_grid = typing_pres.create_preferences_grid()
        self.stack.add_named(self.typing_echo_grid, "typing_echo")
        self._add_navigation_row("typing_echo", self.typing_echo_grid.get_title())

        flat_review_pres = flat_review_presenter.get_presenter()
        self.flat_review_grid = flat_review_pres.create_preferences_grid()
        self.stack.add_named(self.flat_review_grid, "flat_review")
        self._add_navigation_row("flat_review", self.flat_review_grid.get_title())

        mouse_pres = mouse_presenter.get_presenter()
        self.mouse_grid = mouse_pres.create_preferences_grid()
        self.stack.add_named(self.mouse_grid, "mouse")
        self._add_navigation_row("mouse", self.mouse_grid.get_title())

        doc_presenter = document_presenter.get_presenter()
        self.document_grid = doc_presenter.create_preferences_grid(update_title)
        self.stack.add_named(self.document_grid, "documents")
        self._add_navigation_row("documents", self.document_grid.get_title())

        say_all_pres = say_all_presenter.get_presenter()
        self.say_all_grid = say_all_pres.create_preferences_grid()
        self.stack.add_named(self.say_all_grid, "say_all")
        self._add_navigation_row("say_all", self.say_all_grid.get_title())

        pronunciation_manager = pronunciation_dictionary_manager.get_manager()
        self.pronunciation_grid = pronunciation_manager.create_preferences_grid(self.script)
        self.stack.add_named(self.pronunciation_grid, "pronunciation")
        self._add_navigation_row("pronunciation", self.pronunciation_grid.get_title())

        spellcheck_pres = spellcheck_presenter.get_presenter()
        self.spellcheck_grid = spellcheck_pres.create_preferences_grid()
        self.stack.add_named(self.spellcheck_grid, "spellcheck")
        self._add_navigation_row("spellcheck", self.spellcheck_grid.get_title())

        chat_pres = chat_presenter.get_presenter()
        self.chat_grid = chat_pres.create_preferences_grid()
        self.stack.add_named(self.chat_grid, "chat")
        self._add_navigation_row("chat", self.chat_grid.get_title())

        text_attr_mgr = text_attribute_manager.get_manager()
        self.text_attributes_grid = text_attr_mgr.create_preferences_grid()
        self.stack.add_named(self.text_attributes_grid, "text_attributes")
        self._add_navigation_row(
            "text_attributes",
            self.text_attributes_grid.get_title(),
        )

        system_info_presenter = system_information_presenter.get_presenter()
        self.time_and_date_grid = system_info_presenter.create_time_and_date_preferences_grid()
        self.stack.add_named(self.time_and_date_grid, "time_and_date")
        self._add_navigation_row("time_and_date", self.time_and_date_grid.get_title())

        if not self._app_name:
            self.extensions_grid = (
                extension_loader_preferences_grid.ExtensionLoaderPreferencesGrid()
            )
            self.stack.add_named(self.extensions_grid, "extensions")
            self._add_navigation_row("extensions", self.extensions_grid.get_title())

        registry.set_ignore_runtime(False)

        self._page_to_grid: dict[str, preferences_grid_base.PreferencesGridBase] = {
            "speech": self.speech_grid,
            "braille": self.braille_grid,
            "keybindings": self.keybindings_grid,
            "typing_echo": self.typing_echo_grid,
            "flat_review": self.flat_review_grid,
            "say_all": self.say_all_grid,
            "spellcheck": self.spellcheck_grid,
            "chat": self.chat_grid,
            "mouse": self.mouse_grid,
            "documents": self.document_grid,
            "pronunciation": self.pronunciation_grid,
            "sound": self.sound_grid,
            "time_and_date": self.time_and_date_grid,
            "text_attributes": self.text_attributes_grid,
            "profiles": self.profiles_grid,
        }
        if not self._app_name:
            self._page_to_grid["sleep_mode"] = self.sleep_mode_grid
            self._page_to_grid["extensions"] = self.extensions_grid

        for grid in self._page_to_grid.values():
            grid.set_focus_sidebar_callback(self._focus_sidebar)

        first_row = self.listbox.get_row_at_index(0)
        if first_row:
            self.listbox.select_row(first_row)

        msg = "PREFERENCES: Initializing UI complete"
        debug.print_message(debug.LEVEL_ALL, msg, True)

    def _apply_decoration_layout(self) -> None:
        """Syncs headerbar button placement and app icon with the system layout."""

        gtk_settings = Gtk.Settings.get_default()  # pylint: disable=no-value-for-parameter
        if gtk_settings is None:
            return

        layout = gtk_settings.get_property("gtk-decoration-layout") or ":minimize,maximize,close"
        left_layout, _, right_layout = layout.partition(":")

        self.left_headerbar.set_decoration_layout(f"{left_layout}:")
        self.panel_headerbar.set_decoration_layout(f":{right_layout}")

        parent = self._app_icon.get_parent()
        if parent:
            parent.remove(self._app_icon)

        if "close" in left_layout:
            self.panel_headerbar.pack_end(self._app_icon)
        else:
            self.left_headerbar.pack_start(self._app_icon)
        self._app_icon.show()

        if self.get_visible():
            GLib.idle_add(self._sync_headerbar_widths)

    def _sync_headerbar_widths(self) -> None:
        headerbar_width = self.left_headerbar.get_allocated_width()
        sidebar_width = self.sidebar_vbox.get_allocated_width()
        max_width = max(headerbar_width, sidebar_width)

        self.left_headerbar.set_size_request(max_width, -1)
        self.sidebar_vbox.set_size_request(max_width, -1)

    def _set_height_from_sidebar(self) -> None:
        """Set window height to fit the sidebar items, capped at 800px."""

        sidebar_height = self.listbox.get_preferred_height()[1]
        headerbar_height = self.left_headerbar.get_allocated_height()
        height = min(sidebar_height + headerbar_height, 800)
        self.resize(self.get_allocated_width(), height)

    def _on_listbox_key_press(self, _widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
        if event.keyval == Gdk.KEY_Tab:
            self.stack.child_focus(Gtk.DirectionType.TAB_FORWARD)
            return True
        if event.keyval == Gdk.KEY_ISO_Left_Tab:
            # Shift+Tab from listbox should go back to menu button
            self.menu_button.grab_focus()
            return True
        return False

    def _focus_sidebar(self) -> None:
        """Move focus to the sidebar navigation listbox."""

        selected_row = self.listbox.get_selected_row()
        if selected_row:
            selected_row.grab_focus()
        else:
            self.listbox.grab_focus()

    def _add_navigation_row(self, panel_id: str, label_text: str) -> NavigationRow:
        """Add a navigation row to the sidebar listbox and return it."""

        row = NavigationRow(panel_id=panel_id)

        label = Gtk.Label(label=label_text, xalign=0)
        label.set_margin_start(12)
        label.set_margin_end(12)
        label.set_margin_top(6)
        label.set_margin_bottom(6)
        row.add(label)

        self.listbox.add(row)
        return row

    def _on_row_selected(self, _listbox: Gtk.ListBox, row: Gtk.ListBoxRow | None) -> None:
        """Handle listbox row selection and switch to the selected panel."""

        if row is None or not isinstance(row, NavigationRow):
            return

        # Skip if this is a non-selectable header row
        if row.panel_id is None:
            return

        panel_id = row.panel_id

        # Update the panel title in headerbar and accessible name
        if panel_id == "application":
            title = AXObject.get_name(self.script.app)
        else:
            child = self.stack.get_child_by_name(panel_id)
            title = child.get_title()

        self._set_page_title(title)

        self.stack.set_visible_child_name(panel_id)

        # Only reset multi-page grids when switching to a different panel.
        # This preserves sub-category state when focus returns to the
        # sidebar via Shift-Tab without changing the selected row.
        child = self.stack.get_child_by_name(panel_id)
        if isinstance(child, preferences_grid_base.PreferencesGridBase):
            if panel_id != self._current_panel_id:
                child.on_becoming_visible()

        self._current_panel_id = panel_id

    def _init_gui_state(self, include_profiles: bool = False) -> None:
        """Adjust the settings of the various widgets based on user settings."""

        self._reload_all_grids(include_profiles=include_profiles)

    def show_gui(self) -> None:
        """Show the Orca configuration GUI window."""

        def enable_configuring_mode():
            focus_manager.get_manager().set_in_preferences_window(True)
            return False

        if OrcaSetupGUI.WINDOW is not None:
            OrcaSetupGUI.WINDOW.present()
            return

        OrcaSetupGUI.WINDOW = self

        accel_group = Gtk.AccelGroup()
        OrcaSetupGUI.WINDOW.add_accel_group(accel_group)

        # Select first row and set visible child before showing dialog
        first_row = self.listbox.get_row_at_index(0)
        if first_row and isinstance(first_row, NavigationRow) and first_row.panel_id:
            self.listbox.select_row(first_row)
            panel_id = first_row.panel_id
            # Set initial panel title in headerbar
            if panel_id == "application":
                title = AXObject.get_name(self.script.app)
            else:
                child = self.stack.get_child_by_name(panel_id)
                title = child.get_title()
            self._set_page_title(title)
            self.stack.set_visible_child_name(panel_id)

        self.suspend_events("Window being shown.")
        OrcaSetupGUI.WINDOW.show_all()
        self._sync_headerbar_widths()
        self._set_height_from_sidebar()
        self.stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.stack.set_transition_duration(150)
        OrcaSetupGUI.WINDOW.present_with_time(time.time())
        GLib.timeout_add(500, self.resume_events)

        # Set accessible name after window is realized
        def set_accessible_name() -> bool:
            self.update_menu_labels()
            return False

        GLib.idle_add(set_accessible_name)

        # Enable configuring mode after a brief delay to allow initial setup to complete
        GLib.timeout_add(100, enable_configuring_mode)

    def close_for_shutdown(self) -> None:
        """Close preferences UI without prompting during Orca shutdown."""

        msg = "PREFERENCES: Closing preferences UI for shutdown"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        for window in Gtk.Window.list_toplevels():
            parent = window.get_transient_for()
            while parent is not None:
                if parent is self:
                    if isinstance(window, Gtk.Dialog) and window.get_visible():
                        window.response(Gtk.ResponseType.DELETE_EVENT)
                    break
                parent = parent.get_transient_for()

        self.destroy()

    def help_button_clicked(self, _widget: Gtk.Button) -> None:
        """Handle Help button click to show preferences help."""

        learn_mode_presenter.get_presenter().show_help(page="preferences")

    def _reload_all_grids(self, include_profiles: bool = False) -> None:
        """Reload all preference grids from settings."""

        registry = gsettings_registry.get_registry()
        registry.set_ignore_runtime(True)
        for grid in self._page_to_grid.values():
            if grid is self.profiles_grid and not include_profiles:
                continue
            grid.reload()
        registry.set_ignore_runtime(False)

    def apply_button_clicked(self, _widget: Gtk.Button) -> None:
        """Handle Apply button click to save and apply preferences."""

        msg = "PREFERENCES: Apply button clicked"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        save_app = self._app_name or ""

        # Save profiles first so any renames happen before saving other settings
        # (profiles are global, not saved for app-specific preferences)
        if not self._app_name:
            self.profiles_grid.save_settings()
            self._profile_name = profile_manager.get_manager().get_active_profile()

        for grid in self._page_to_grid.values():
            if grid is not self.profiles_grid:
                grid.save_settings(self._profile_name, save_app)

        orca.load_user_settings(self.script, skip_reload_message=True)

        # Speak the settings reloaded message after a delay to ensure speech has fully started.
        def speak_settings_reloaded() -> bool:
            presentation_manager.get_manager().speak_message(messages.SETTINGS_RELOADED)
            return False

        GLib.timeout_add(100, speak_settings_reloaded)

        self._reload_all_grids()

        # Re-apply user keybinding overrides and refresh grabs so changes work immediately.
        command_manager.get_manager().activate_commands("Applied preferences")

        msg = "PREFERENCES: Handling Apply button click complete"
        debug.print_message(debug.LEVEL_ALL, msg, True)

    def _revert_unsaved_changes(self) -> None:
        """Discard in-session overrides and restore live state from dconf."""

        gsettings_registry.get_registry().clear_runtime_values()
        orca.load_user_settings(self.script, skip_reload_message=True)
        for grid in self._page_to_grid.values():
            grid.revert_changes()

    def _maybe_prompt_about_profile_switch(self) -> None:
        """If the active profile was switched this session, ask whether to keep it."""

        current_profile = profile_manager.get_manager().get_active_profile()
        if current_profile == self._original_profile:
            return

        current_label = self._get_current_profile_label()
        original_label = self.profiles_grid.get_profile_label(self._original_profile)
        dialog = Gtk.MessageDialog(
            transient_for=self,
            modal=True,
            message_type=Gtk.MessageType.QUESTION,
            buttons=Gtk.ButtonsType.NONE,
            text=guilabels.PREFERENCES_PROFILE_SWITCHED,
        )
        use_button = dialog.add_button(
            guilabels.PROFILE_USE % current_label,
            Gtk.ResponseType.YES,
        )
        use_button.get_style_context().add_class("suggested-action")
        dialog.add_button(
            guilabels.PROFILE_SWITCH_BACK_TO % original_label,
            Gtk.ResponseType.NO,
        )

        dialog.show_all()
        dialog.present()
        use_button.grab_focus()
        response = dialog.run()
        dialog.destroy()

        if response == Gtk.ResponseType.NO:
            profile_manager.get_manager().load_profile(self._original_profile)

    def cancel_button_clicked(self, _widget: Gtk.Button) -> None:
        """Handle Cancel button click to close window without saving."""

        msg = "PREFERENCES: Cancel button clicked"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        self._revert_unsaved_changes()
        self._maybe_prompt_about_profile_switch()
        self.destroy()

        msg = "PREFERENCES: Handling Cancel button click complete"
        debug.print_message(debug.LEVEL_ALL, msg, True)

    def ok_button_clicked(self, widget: Gtk.Button | None = None) -> None:
        """Handle OK button click to save preferences and close window."""

        msg = "PREFERENCES: OK button clicked"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        self.apply_button_clicked(widget)  # type: ignore[arg-type]
        self.destroy()

        msg = "PREFERENCES: Handling OK button click complete"
        debug.print_message(debug.LEVEL_ALL, msg, True)

    def _on_save_profile_as(self) -> None:
        """Handle Save Profile As menu item."""

        new_profile = self.profiles_grid.get_new_profile_name()
        if new_profile is None:
            return

        self._profile_name = new_profile[1]
        self.apply_button_clicked(None)
        self._on_profile_loaded(new_profile)

    def window_closed(self, _widget: Gtk.Widget, _event: Any) -> bool:
        """Handle window close signal by suspending events temporarily."""

        msg = "PREFERENCES: Window is being closed"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        has_unsaved_changes = self._has_unsaved_changes(
            include_profiles=not self._app_name,
        )

        # Check if profile was switched during this session
        current_profile = profile_manager.get_manager().get_active_profile()
        profile_was_switched = current_profile != self._original_profile

        if has_unsaved_changes:
            dialog = Gtk.MessageDialog(
                transient_for=self,
                modal=True,
                message_type=Gtk.MessageType.QUESTION,
                buttons=Gtk.ButtonsType.NONE,
                text=guilabels.PREFERENCES_CLOSE_WITHOUT_SAVE,
            )
            dialog.format_secondary_text(guilabels.PREFERENCES_CHANGES_WILL_BE_LOST)

            profile_label = self._get_current_profile_label()
            save_button = dialog.add_button(
                guilabels.MENU_SAVE_PROFILE % profile_label,
                Gtk.ResponseType.YES,
            )
            save_button.get_style_context().add_class("suggested-action")

            # Only show "New Profile" for global preferences (profiles are global)
            if not self._app_name:
                dialog.add_button(guilabels.PROFILE_CREATE_NEW, Gtk.ResponseType.APPLY)

            dialog.add_button(guilabels.BTN_CLOSE_WITHOUT_SAVING, Gtk.ResponseType.NO)
            dialog.add_button(guilabels.DIALOG_CANCEL, Gtk.ResponseType.CANCEL)

            dialog.show_all()
            dialog.present()
            save_button.grab_focus()
            response = dialog.run()
            dialog.destroy()

            if response == Gtk.ResponseType.YES:
                self.apply_button_clicked(None)
            elif response == Gtk.ResponseType.APPLY:
                self._on_save_profile_as()
            elif response in (Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT):
                return True
            else:
                self._revert_unsaved_changes()

        elif profile_was_switched:
            self._maybe_prompt_about_profile_switch()

        self.suspend_events("Window being closed.")
        GObject.timeout_add(1000, self.resume_events)

        msg = "PREFERENCES: Window closure complete"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        return False

    def window_destroyed(self, _widget: Gtk.Widget) -> None:
        """Handle window destroyed signal by clearing window reference."""

        msg = "PREFERENCES WINDOW: Window is being destroyed"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        OrcaSetupGUI.WINDOW = None

        msg = "PREFERENCES WINDOW: Window destruction complete"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        focus_manager.get_manager().set_in_preferences_window(False)

    def _has_unsaved_changes(self, include_profiles: bool = False) -> bool:
        """Returns True if any preference grid has unsaved changes."""

        for name, grid in self._page_to_grid.items():
            if grid is self.profiles_grid and not include_profiles:
                continue
            if grid.has_changes():
                msg = f"PREFERENCES: Grid '{name}' has unsaved changes"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True
        return False

    def resume_events(self, reason: str = "") -> bool:
        """Re-register event listeners suspended during UI creation and teardown."""

        msg = f"PREFERENCES: Re-registering events. {reason}"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        manager = event_manager.get_manager()
        for event in self._EVENTS_TO_SUSPEND:
            manager.register_listener(event)
        return False

    def suspend_events(self, reason: str = "") -> None:
        """Deregister event listeners that flood Orca during UI creation and teardown."""

        msg = f"PREFERENCES: Suspending events. {reason}"
        debug.print_message(debug.LEVEL_ALL, msg, True)

        manager = event_manager.get_manager()
        for event in self._EVENTS_TO_SUSPEND:
            manager.deregister_listener(event)

    def _get_current_profile_label(self) -> str:
        """Get the display label for the current profile, including pending renames."""

        return self.profiles_grid.get_current_profile_label()

    def _get_panel_title(self, page_title: str) -> str:
        """Build the panel header title with profile in parentheses."""

        profile_label = self._get_current_profile_label()
        return f"{page_title} ({profile_label})"

    def _set_page_title(self, title: str) -> None:
        """Set the current page title, updating panel header and accessible name."""

        self._current_page_title = title
        self.panel_headerbar.set_title(self._get_panel_title(title))
        self.get_accessible().set_name(self._get_accessible_name(title))

    def _get_accessible_name(self, page_title: str = "") -> str:
        """Build the accessible name for the window."""

        if self._app_name:
            base_title = guilabels.PREFERENCES_APPLICATION_TITLE % self._app_name
        else:
            base_title = guilabels.DIALOG_SCREEN_READER_PREFERENCES_ACCESSIBLE

        profile_label = self._get_current_profile_label()

        if page_title:
            return f"{base_title}, {page_title}, {profile_label}"
        return f"{base_title}, {profile_label}"

    def update_menu_labels(self) -> None:
        """Update Apply and Save menu items and panel title to show the current profile."""

        profile_label = self._get_current_profile_label()
        self.apply_item.set_property("text", guilabels.MENU_APPLY_PROFILE % profile_label)
        self.save_item.set_property("text", guilabels.MENU_SAVE_PROFILE % profile_label)

        if self._app_name:
            self.left_headerbar.set_subtitle(self._app_name)
        else:
            self.left_headerbar.set_subtitle(None)

        if self._current_page_title:
            self._set_page_title(self._current_page_title)

    def _on_profile_loaded(self, profile: list[str]) -> None:
        """Handle profile loaded callback and reload all preference grids."""

        if not self.get_realized():
            return

        self._profile_name = profile[1]
        self._init_gui_state(include_profiles=False)
        self.update_menu_labels()



# Orca
#
# Copyright (C) 2011-2013 Igalia, S.L.
#
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-locals
# pylint: disable=too-many-statements

"""Heuristic means to infer the functional/displayed label of a widget."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import ax_cache_manager, debug
from .ax_component import AXComponent
from .ax_hypertext import AXHypertext
from .ax_object import AXObject
from .ax_table import AXTable
from .ax_text import AXText
from .ax_utilities import AXUtilities

if TYPE_CHECKING:
    from collections.abc import Hashable

    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .scripts import default


class _LabelInferenceCache:
    """Provides label-inference access to manager-backed cached values."""

    LINE_CONTENTS = "LabelInference.line-contents"
    TEXT_EXTENTS = "LabelInference.text-extents"
    IS_WIDGET = "LabelInference.is-widget"

    def __init__(self) -> None:
        self._manager = ax_cache_manager.get_manager()
        for namespace in (
            self.LINE_CONTENTS,
            self.TEXT_EXTENTS,
            self.IS_WIDGET,
        ):
            self._manager.register_cache(
                self,
                namespace,
                lifetime=ax_cache_manager.Lifetime.OWNER,
                clear_on_demand=ax_cache_manager.ClearPolicy.PRESERVE,
                clear_interval_seconds=None,
            )
        self._line_contents_cache = self._manager.get_cache(self, self.LINE_CONTENTS)
        self._text_extents_cache = self._manager.get_cache(self, self.TEXT_EXTENTS)
        self._is_widget_cache = self._manager.get_cache(self, self.IS_WIDGET)

    def clear(self, reason: str) -> None:
        """Clears cached label-inference values."""

        for cache in (
            self._line_contents_cache,
            self._text_extents_cache,
            self._is_widget_cache,
        ):
            if cache is not None:
                cache.invalidate(reason)

    def get_line_contents(
        self, obj: Atspi.Accessible
    ) -> list[tuple[Atspi.Accessible, int, int, str]] | None:
        """Returns cached line contents for obj."""

        if self._line_contents_cache is None:
            return None

        contents = self._line_contents_cache.get(
            ax_cache_manager.get_object_key(obj), ax_cache_manager.MISSING
        )
        if contents is ax_cache_manager.MISSING:
            return None
        return list(contents)

    def set_line_contents(
        self,
        obj_key: Hashable,
        contents: list[tuple[Atspi.Accessible, int, int, str]],
    ) -> None:
        """Stores line contents for obj."""

        if self._line_contents_cache is not None:
            self._line_contents_cache.put(obj_key, list(contents))

    def get_text_extents(
        self,
        obj: Atspi.Accessible,
        start_offset: int,
        end_offset: int,
    ) -> tuple[int, int, int, int] | None:
        """Returns cached text extents for obj."""

        key = ax_cache_manager.get_object_key(obj), start_offset, end_offset
        if self._text_extents_cache is None:
            return None

        return self._text_extents_cache.get(key, None)

    def set_text_extents(
        self,
        obj: Atspi.Accessible,
        start_offset: int,
        end_offset: int,
        extents: tuple[int, int, int, int],
    ) -> None:
        """Stores text extents for obj."""

        key = ax_cache_manager.get_object_key(obj), start_offset, end_offset
        if self._text_extents_cache is not None:
            self._text_extents_cache.put(key, extents)

    def get_is_widget(self, obj: Atspi.Accessible) -> bool | None:
        """Returns the cached widget decision for obj."""

        if self._is_widget_cache is None:
            return None

        return self._is_widget_cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_is_widget(self, obj: Atspi.Accessible, is_widget: bool) -> None:
        """Stores the widget decision for obj."""

        if self._is_widget_cache is not None:
            self._is_widget_cache.put(ax_cache_manager.get_object_key(obj), is_widget)


class LabelInference:
    """Heuristic means to infer the functional/displayed label of a widget."""

    def __init__(self, script: default.Script) -> None:
        self._script = script
        self._cache = _LabelInferenceCache()

    def infer(
        self,
        obj: Atspi.Accessible,
        focused_only: bool = True,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Attempt to infer the functional/displayed label of obj."""

        tokens = ["LABEL INFERENCE: Infer label for", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if not obj:
            return None, []

        if focused_only and not AXUtilities.is_focused(obj):
            tokens = ["LABEL INFERENCE:", obj, "is not focused"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None, []

        result: str | None = None
        objects: list[Atspi.Accessible] = []
        if not result:
            result, objects = self._infer_from_text_left(obj)
            debug.print_message(debug.LEVEL_INFO, f"LABEL INFERENCE: Text Left: '{result}'", True)
        if not result or self._prefer_right(obj):
            temp_result = self._infer_from_text_right(obj)
            if temp_result[0] is not None:
                result, objects = temp_result
            debug.print_message(debug.LEVEL_INFO, f"LABEL INFERENCE: Text Right: '{result}'", True)
        if not result:
            result, objects = self._infer_from_table(obj)
            debug.print_message(debug.LEVEL_INFO, f"LABEL INFERENCE: Table: '{result}'", True)
        if not result:
            result, objects = self._infer_from_text_above(obj)
            debug.print_message(debug.LEVEL_INFO, f"LABEL INFERENCE: Text Above: '{result}'", True)
        if not result:
            result, objects = self._infer_from_text_below(obj)
            debug.print_message(debug.LEVEL_INFO, f"LABEL INFERENCE: Text Below: '{result}'", True)

        # TODO - We probably do not wish to "infer" from these. Instead, we
        # should ensure that this content gets presented as part of the widget.
        # (i.e. the label is something on screen. Widget name and description
        # are each something other than a label.)
        if not result:
            result, objects = AXObject.get_name(obj), []
            debug.print_message(debug.LEVEL_INFO, f"LABEL INFERENCE: Name: '{result}'", True)

        if result:
            result = result.strip()
            result = result.replace("\n", " ")

        # Desperate times call for desperate measures....
        if not result:
            result, objects = self._infer_from_text_left(obj, proximity=200)
            debug.print_message(
                debug.LEVEL_INFO,
                f"LABEL INFERENCE: Text Left with proximity of 200: '{result}'",
                True,
            )

        self._clear_cache()
        return result, objects

    def _clear_cache(self) -> None:
        """Clears values stored for one label inference."""

        self._cache.clear("label inference complete")

    def _prefer_right(self, obj: Atspi.Accessible) -> bool:
        """Returns True if we should prefer text on the right."""

        return AXUtilities.is_check_box(obj) or AXUtilities.is_radio_button(obj)

    def _prevent_right(self, obj: Atspi.Accessible) -> bool:
        """Returns True if we should not permit inference based on text to the right of obj."""

        return AXUtilities.is_combo_box(obj) or AXUtilities.is_list_box(obj)

    def _prefer_top(self, obj: Atspi.Accessible) -> bool:
        """Returns True if we should prefer text above, rather than below for obj."""

        return AXUtilities.is_combo_box(obj) or AXUtilities.is_list_box(obj)

    def _prevent_below(self, obj: Atspi.Accessible) -> bool:
        """Returns True if we should not permit inference based on text below obj."""

        return not AXUtilities.is_text_input(obj)

    def _is_simple_object(self, obj: Atspi.Accessible | None) -> bool:
        """Returns True if the given object has 'simple' contents."""

        if obj is None:
            return False

        def is_match(x: Atspi.Accessible) -> bool:
            return AXUtilities.is_web_element(x) and not AXUtilities.is_link(x)

        children = list(AXObject.iter_children(obj, is_match))
        if len(children) > 1:
            return False

        string = AXText.get_all_text(obj).strip()
        return string.count("\ufffc") <= 1

    def _cannot_label(self, obj: Atspi.Accessible | None) -> bool:
        """Returns True if the given object should not be treated as a label."""

        if obj is None:
            return True

        if AXUtilities.is_heading(obj) or AXUtilities.is_list_item(obj):
            return True

        return self._is_widget(obj)

    def _is_widget(self, obj: Atspi.Accessible | None) -> bool:
        """Returns True if the given object is a widget."""

        if obj is None:
            return False

        rv = self._cache.get_is_widget(obj)
        if rv is not None:
            return rv

        is_widget = AXUtilities.is_widget(obj) or AXUtilities.is_menu_related(obj)
        if not is_widget and AXUtilities.is_editable(obj):
            is_widget = True

        self._cache.set_is_widget(obj, is_widget)
        return is_widget

    def _get_extents(
        self,
        obj: Atspi.Accessible | None,
        start_offset: int = 0,
        end_offset: int = -1,
    ) -> tuple[int, int, int, int]:
        """Returns (x, y, width, height) of the text at the given offsets."""

        if obj is None:
            return 0, 0, 0, 0

        rv = self._cache.get_text_extents(obj, start_offset, end_offset)
        if rv:
            return rv

        extents = 0, 0, 0, 0
        if AXObject.supports_text(obj):
            if not AXUtilities.is_text_input(obj):
                if end_offset == -1:
                    end_offset = AXText.get_character_count(obj)
                rect = AXText.get_range_rect(obj, start_offset, end_offset)
                extents = rect.x, rect.y, rect.width, rect.height

        if not (extents[2] and extents[3]):
            ext = AXComponent.get_rect(obj)
            extents = ext.x, ext.y, ext.width, ext.height

        self._cache.set_text_extents(obj, start_offset, end_offset, extents)
        return extents

    def _create_label_from_contents(
        self,
        obj: Atspi.Accessible,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Gets the functional label text associated with the object obj."""

        if not self._is_simple_object(obj):
            return None, []

        if self._cannot_label(obj):
            return None, []

        contents = self._script.utilities.get_object_contents_at_offset(obj, use_cache=False)
        objects = [content[0] for content in contents]
        if list(filter(self._is_widget, objects)):
            return None, []

        strings = [content[3] for content in contents]
        return "".join(strings), objects

    def _get_line_contents(
        self,
        obj: Atspi.Accessible,
        start: int = 0,
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Get the (obj, start_offset, end_offset, string) tuples for the line containing obj."""

        rv = self._cache.get_line_contents(obj)
        if rv:
            return rv

        key = ax_cache_manager.get_object_key(obj)
        if self._is_widget(obj):
            start = AXHypertext.get_link_start_offset(obj)
            obj = AXObject.get_parent(obj)

        rv = self._script.utilities.get_line_contents_at_offset(obj, start, True, False)
        if rv is None:
            rv = []
        self._cache.set_line_contents(key, rv)

        return rv

    def _infer_from_text_left(
        self,
        obj: Atspi.Accessible,
        proximity: int = 75,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Attempt to infer the functional/displayed label of obj from contents on its left."""

        extents = self._get_extents(obj)
        contents = self._get_line_contents(obj)
        content = [o for o in contents if o[0] == obj]
        try:
            index = contents.index(content[0])
        except IndexError:
            index = len(contents)

        on_left = contents[0:index]
        start = 0
        for i in range(len(on_left) - 1, -1, -1):
            left_obj, _left_start, _left_end, _left_string = on_left[i]
            left_extents = self._get_extents(left_obj)
            if left_extents[0] > extents[0] or self._cannot_label(left_obj):
                start = i + 1
                break

        on_left = on_left[start:]
        if not (on_left and on_left[0]):
            return None, []

        left_obj, start, end, _string = on_left[-1]
        left_extents = self._get_extents(left_obj, start, end)
        distance = extents[0] - (left_extents[0] + left_extents[2])
        if 0 <= distance <= proximity:
            strings = [content[3] for content in on_left]
            result = "".join(strings).strip()
            if result:
                return result, [content[0] for content in on_left]

        return None, []

    def _infer_from_text_right(
        self,
        obj: Atspi.Accessible,
        proximity: int = 25,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Attempt to infer the functional/displayed label of obj from contents on its right."""

        if self._prevent_right(obj):
            return None, []

        extents = self._get_extents(obj)
        contents = self._get_line_contents(obj)
        content = [o for o in contents if o[0] == obj]
        try:
            index = contents.index(content[0])
        except IndexError:
            index = len(contents)

        on_right = contents[min(len(contents), index + 1) :]
        end = len(on_right)
        for i, item in enumerate(on_right):
            if self._cannot_label(item[0]):
                if not self._prefer_right(obj):
                    return None, []
                end = i + 1
                break

        on_right = on_right[0:end]
        if not (on_right and on_right[0]):
            return None, []

        right_obj, start, end, _string = on_right[0]
        right_extents = self._get_extents(right_obj, start, end)
        distance = right_extents[0] - (extents[0] + extents[2])
        if distance <= proximity or self._prefer_right(obj):
            strings = [content[3] for content in on_right]
            result = "".join(strings).strip()
            if result:
                return result, [content[0] for content in on_right]

        return None, []

    def _infer_from_text_above(
        self,
        obj: Atspi.Accessible,
        proximity: int = 20,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Attempt to infer the functional/displayed label of obj from contents above it."""

        this_line = self._get_line_contents(obj)
        content = [o for o in this_line if o[0] == obj]
        try:
            index = this_line.index(content[0])
        except IndexError:
            return None, []
        if index > 0:
            return None, []

        prev_obj, prev_offset = self._script.utilities.previous_context(
            this_line[0][0],
            this_line[0][1],
            True,
        )
        prev_line = self._get_line_contents(prev_obj, prev_offset)
        if len(prev_line) != 1:
            return None, []

        prev_obj, start, end, string = prev_line[0]
        if self._cannot_label(prev_obj):
            return None, []

        if string.endswith("\n"):
            string = string[:-1]
            end -= 1

        if string.strip():
            x, y, _width, height = self._get_extents(prev_obj, start, end)
            obj_x, obj_y, _obj_width, _obj_height = self._get_extents(obj)
            distance = obj_y - (y + height)
            if 0 <= distance <= proximity and x <= obj_x:
                return string.strip(), [prev_obj]

        return None, []

    def _infer_from_text_below(
        self,
        obj: Atspi.Accessible,
        proximity: int = 20,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Attempt to infer the functional/displayed label of obj from contents below it."""

        if self._prevent_below(obj):
            return None, []

        this_line = self._get_line_contents(obj)
        content = [o for o in this_line if o[0] == obj]
        try:
            index = this_line.index(content[0])
        except IndexError:
            return None, []
        if index > 0:
            return None, []

        next_obj, next_offset = self._script.utilities.next_context(
            this_line[-1][0],
            this_line[-1][2] - 1,
            True,
        )
        next_line = self._get_line_contents(next_obj, next_offset)
        if len(next_line) != 1 or self._cannot_label(next_line[0][0]):
            return None, []

        next_obj, start, end, string = next_line[0]

        if string.strip():
            _x, y, _width, _height = self._get_extents(next_obj, start, end)
            _obj_x, obj_y, _obj_width, obj_height = self._get_extents(obj)
            distance = y - (obj_y + obj_height)
            if 0 <= distance <= proximity:
                return string.strip(), [next_obj]

        return None, []

    def _choose_vertical_label(
        self,
        obj: Atspi.Accessible,
        cell_above: Atspi.Accessible | None,
        cell_below: Atspi.Accessible | None,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Returns the best label from cells above or below obj."""

        label_above: str | None = None
        sources_above: list[Atspi.Accessible] = []
        if cell_above:
            label_above, sources_above = self._create_label_from_contents(cell_above)
            if label_above and self._prefer_top(obj):
                return label_above.strip(), sources_above

        label_below: str | None = None
        sources_below: list[Atspi.Accessible] = []
        if cell_below and not self._prevent_below(obj):
            label_below, sources_below = self._create_label_from_contents(cell_below)

        if label_above and label_below:
            _obj_x, obj_y, _obj_width, obj_height = self._get_extents(obj)
            _above_x, above_y, _above_width, above_height = self._get_extents(cell_above)
            _below_x, below_y, _below_width, _below_height = self._get_extents(cell_below)
            delta_above = obj_y - (above_y + above_height)
            delta_below = below_y - (obj_y + obj_height)
            if delta_above <= delta_below:
                return label_above.strip(), sources_above
            return label_below.strip(), sources_below

        if label_above:
            return label_above.strip(), sources_above
        if label_below:
            return label_below.strip(), sources_below
        return None, []

    def _infer_from_table_header_row(
        self,
        obj: Atspi.Accessible,
        grid: Atspi.Accessible,
        colindex: int,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Attempts to infer the label from the first row of a table-like grid."""

        columns = AXTable.get_column_count(grid)
        first_row = [AXTable.get_cell_at(grid, 0, i) for i in range(columns)]
        if not first_row or list(filter(self._is_widget, first_row)) or colindex < 0:
            return None, []

        def is_match(x: Atspi.Accessible) -> bool:
            if not AXObject.get_child_count(x):
                return False
            return not AXUtilities.have_same_role(AXObject.get_child(x, 0), obj)

        rows = AXTable.get_row_count(grid)
        cells = [AXTable.get_cell_at(grid, i, colindex) for i in range(1, rows)]
        if list(filter(is_match, cells)):
            return None, []

        label, sources = self._create_label_from_contents(first_row[colindex])
        if label:
            return label.strip(), sources
        return None, []

    def _infer_from_table(
        self,
        obj: Atspi.Accessible,
        proximity_for_right: int = 50,
    ) -> tuple[str | None, list[Atspi.Accessible]]:
        """Attempt to infer the functional/displayed label of obj from neighboring cells."""

        cell = AXUtilities.find_ancestor(obj, AXUtilities.is_table_cell)
        parent = AXObject.get_parent(obj)
        if not self._is_simple_object(cell) or cell not in [parent, AXObject.get_parent(parent)]:
            return None, []

        grid = AXUtilities.find_ancestor(cell, AXUtilities.is_table)
        if not grid:
            return None, []

        cell_left = cell_right = cell_above = cell_below = None
        gridrow = AXUtilities.find_ancestor(cell, AXUtilities.is_table_row)
        rowindex, colindex = AXTable.get_cell_coordinates(cell, prefer_attribute=False)
        if colindex > -1:
            cell_left = AXTable.get_cell_at(grid, rowindex, colindex - 1)
            cell_right = AXTable.get_cell_at(grid, rowindex, colindex + 1)
            cell_above = AXTable.get_cell_at(grid, rowindex - 1, colindex)
            cell_below = AXTable.get_cell_at(grid, rowindex + 1, colindex)
        elif gridrow and AXObject.get_parent(cell) == gridrow:
            cellindex = AXObject.get_index_in_parent(cell)
            cell_left = AXObject.get_child(gridrow, cellindex - 1)
            cell_right = AXObject.get_child(gridrow, cellindex + 1)
            rowindex = AXObject.get_index_in_parent(gridrow)
            gridrow_parent = AXObject.get_parent(gridrow)
            if rowindex > 0:
                row_above = AXObject.get_child(gridrow_parent, rowindex - 1)
                cell_above = AXObject.get_child(row_above, cellindex)
            if rowindex + 1 < AXObject.get_child_count(grid):
                row_below = AXObject.get_child(gridrow_parent, rowindex + 1)
                cell_below = AXObject.get_child(row_below, cellindex)

        if cell_left and not self._prefer_right(obj):
            label, sources = self._create_label_from_contents(cell_left)
            if label:
                return label.strip(), sources

        if cell_right and not self._prevent_right(obj):
            obj_x, _obj_y, obj_width, _obj_height = self._get_extents(obj)
            x, _y, _width, _height = self._get_extents(cell_right)
            distance = x - (obj_x + obj_width)
            if distance <= proximity_for_right or self._prefer_right(obj):
                label, sources = self._create_label_from_contents(cell_right)
                if label:
                    return label.strip(), sources

        label, sources = self._choose_vertical_label(obj, cell_above, cell_below)
        if label:
            return label, sources

        return self._infer_from_table_header_row(obj, grid, colindex)



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2006-2009 Brailcom, o.p.s.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=unused-argument
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments

"""Functionality for working with speech servers."""

from __future__ import annotations

from enum import Enum, StrEnum
from typing import TYPE_CHECKING, Any

from . import debug, gsettings_registry, guilabels, language_utilities
from .acss import ACSS
from .ax_utilities_debugging import AXUtilitiesDebugging


class VoiceType(StrEnum):
    """Voice type identifiers. Values match GSettings path components."""

    DEFAULT = "default"
    UPPERCASE = "uppercase"
    HYPERLINK = "hyperlink"
    SYSTEM = "system"


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.PunctuationStyle",
    values={"all": 0, "most": 1, "some": 2, "none": 3},
)
class PunctuationStyle(Enum):
    """Punctuation style enumeration."""

    NONE = 3
    SOME = 2
    MOST = 1
    ALL = 0

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.CapitalizationStyle",
    values={"none": 0, "spell": 1, "icon": 2},
)
class CapitalizationStyle(Enum):
    """Capitalization style enumeration with string values from settings."""

    NONE = "none"
    SPELL = "spell"
    ICON = "icon"

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import ClassVar

    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from . import input_event


class VoiceFamily(dict):
    """Holds the family description for a voice."""

    NAME = "name"
    GENDER = "gender"
    LANG = "lang"
    DIALECT = "dialect"
    VARIANT = "variant"

    MALE = "male"
    FEMALE = "female"

    settings: ClassVar[dict[str, str | None]] = {
        NAME: None,
        GENDER: None,
        LANG: None,
        DIALECT: None,
        VARIANT: None,
    }

    def __init__(self, props: dict[str, Any] | None) -> None:
        """Create and initialize VoiceFamily."""
        dict.__init__(self)

        self.update(VoiceFamily.settings)
        if props:
            self.update(props)


class SayAllContext:
    """Contains information about the current state of a Say All operation."""

    PROGRESS = 0
    INTERRUPTED = 1
    COMPLETED = 2

    def __init__(
        self,
        obj: Atspi.Accessible,
        utterance: str,
        start_offset: int = -1,
        end_offset: int = -1,
    ) -> None:
        self.obj = obj
        self.utterance = utterance
        self.start_offset = start_offset
        self.end_offset = end_offset
        self.current_offset = start_offset
        self.current_end_offset = None

    def __str__(self) -> str:
        obj_string = AXUtilitiesDebugging.as_string(self.obj)
        return (
            f"SAY ALL: {obj_string} '{self.utterance}' ({self.start_offset}-{self.end_offset}, "
            f"current: {self.current_offset})"
        )

    def copy(self) -> SayAllContext:
        """Returns a copy of this SayAllContext."""

        new = SayAllContext(self.obj, self.utterance, self.start_offset, self.end_offset)
        new.current_offset = self.current_offset
        new.current_end_offset = self.current_end_offset
        return new

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SayAllContext):
            return False
        return (
            self.start_offset == other.start_offset
            and self.end_offset == other.end_offset
            and self.obj == other.obj
            and self.utterance == other.utterance
        )

    def __hash__(self) -> int:
        return hash((self.start_offset, self.end_offset, self.obj, self.utterance))


class SpeechServer:
    """Provides speech server abstraction."""

    _LOG_PREFIX: ClassVar[str] = "SPEECH"
    _active_servers: ClassVar[dict[str, SpeechServer]] = {}

    DEFAULT_SERVER_ID = "default"
    _SERVER_NAMES: ClassVar[dict[str, str]] = {
        DEFAULT_SERVER_ID: guilabels.DEFAULT_SYNTHESIZER,
    }

    _ACSS_DEFAULTS: ClassVar[tuple[tuple[str, Any], ...]] = (
        (ACSS.FAMILY, {}),
        (ACSS.RATE, 50),
        (ACSS.AVERAGE_PITCH, 5.0),
        (ACSS.PITCH_RANGE, 5.0),
        (ACSS.GAIN, 10.0),
    )

    def __init__(self, server_id: str = "") -> None:
        self._id = server_id
        self._default_voice: dict[str, Any] = {}
        self._default_voice_name: str = ""
        self._current_voice_properties: dict[str, Any] = {}
        self._current_punctuation_level: PunctuationStyle = PunctuationStyle.MOST

    @staticmethod
    def get_factory_name() -> str:
        """Returns a localized name describing this factory."""

        return ""

    @staticmethod
    def get_speech_servers() -> list[Any]:
        """Returns a list of available speech servers."""

        return []

    @classmethod
    def _get_speech_server(cls, server_id: str) -> SpeechServer | None:
        """Return an active server for given id.

        Attempt to create the server if it doesn't exist yet.  Returns None
        when it is not possible to create the server.
        """

        if server_id not in cls._active_servers:
            cls(server_id)
        return cls._active_servers.get(server_id)

    @classmethod
    def get_speech_server(cls, info: list[str] | None = None) -> SpeechServer | None:
        """Gets a given SpeechServer based upon the [name, id] info."""

        this_id = info[1] if info is not None else cls.DEFAULT_SERVER_ID
        return cls._get_speech_server(this_id)

    @classmethod
    def shutdown_active_servers(cls) -> None:
        """Cleans up and shuts down this factory."""

        servers = list(cls._active_servers.values())
        for server in servers:
            server.shutdown()

    def get_info(self) -> list[str]:
        """Returns [name, id] of the current speech server."""

        return [self._SERVER_NAMES.get(self._id, self._id), self._id]

    def get_voice_families(self) -> list[VoiceFamily]:
        """Returns a list of all known VoiceFamily instances provided by the server."""

        return []

    def _get_default_voice_language(
        self,
        voices: tuple[tuple[str, str, str | None], ...],
    ) -> str:
        """Returns the default language string based on the current locale and available voices."""

        locale_lang, locale_dialect = language_utilities.get_current_language_and_dialect()
        if not locale_dialect:
            return ""

        locale_language = f"{locale_lang}-{locale_dialect}"
        for _name, lang, _variant in voices:
            if lang == locale_language:
                return locale_language
        for _name, lang, _variant in voices:
            if lang == locale_lang:
                return locale_lang
        return locale_language

    def _build_voice_families(
        self,
        voices: tuple[tuple[str, str, str | None], ...],
    ) -> list[VoiceFamily]:
        """Builds VoiceFamily list from raw voice tuples, prepending the default voice."""

        default_lang = self._get_default_voice_language(voices)
        voices = ((self._default_voice_name, default_lang, None), *voices)

        families = []
        for name, lang, variant in voices:
            families.append(
                VoiceFamily(
                    {
                        VoiceFamily.NAME: name,
                        VoiceFamily.LANG: lang.partition("-")[0],
                        VoiceFamily.DIALECT: lang.partition("-")[2],
                        VoiceFamily.VARIANT: variant,
                    },
                ),
            )
        return families

    def get_voice_family(self) -> VoiceFamily:
        """Returns the current voice family as a VoiceFamily dictionary."""

        return VoiceFamily({})

    def set_voice_family(self, family: VoiceFamily) -> None:
        """Sets the voice family to family VoiceFamily dictionary."""

    def speak_character(
        self,
        character: str,
        acss: ACSS | None = None,
        cap_style: CapitalizationStyle | None = None,
    ) -> None:
        """Speaks character."""

    def speak_key_event(self, event: input_event.KeyboardEvent, acss: ACSS | None = None) -> None:
        """Speaks event."""

    def speak(self, text: str | None = None, acss: ACSS | None = None) -> None:
        """Speaks text using the voice specified by acss."""

    def say_all(
        self,
        utterance_iterator: Iterator[tuple[SayAllContext, ACSS]],
        progress_callback: Callable[[SayAllContext, int], None],
    ) -> None:
        """Iterates through the given utterance_iterator, speaking each utterance."""

    def set_default_voice(self, default_voice: dict[str, Any]) -> None:
        """Sets the default voice ACSS properties for fallback use."""

        self._default_voice = default_voice

    def _apply_acss(self, acss: dict[str, Any] | None) -> None:
        """Merges default voice with acss and stores in current voice properties."""

        merged: dict[str, Any] = dict(self._default_voice)
        if acss is not None:
            merged.update(acss)
        for acss_property, default in self._ACSS_DEFAULTS:
            self._current_voice_properties[acss_property] = merged.get(acss_property, default)

    def _change_default_speech_rate(self, step: int, decrease: bool = False) -> None:
        delta = step * (-1 if decrease else 1)
        rate = self._default_voice.get(ACSS.RATE, 50)
        new_rate = max(0, min(100, rate + delta))
        self._default_voice[ACSS.RATE] = new_rate
        self._current_voice_properties[ACSS.RATE] = new_rate
        msg = f"{self._LOG_PREFIX}: Rate set to {new_rate}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _change_default_speech_pitch(self, step: float, decrease: bool = False) -> None:
        delta = step * (-1 if decrease else 1)
        pitch = self._default_voice.get(ACSS.AVERAGE_PITCH, 5)
        new_pitch = max(0, min(10, pitch + delta))
        self._default_voice[ACSS.AVERAGE_PITCH] = new_pitch
        self._current_voice_properties[ACSS.AVERAGE_PITCH] = new_pitch
        msg = f"{self._LOG_PREFIX}: Pitch set to {new_pitch}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _change_default_speech_volume(self, step: float, decrease: bool = False) -> None:
        delta = step * (-1 if decrease else 1)
        volume = self._default_voice.get(ACSS.GAIN, 10)
        new_volume = max(0, min(10, volume + delta))
        self._default_voice[ACSS.GAIN] = new_volume
        self._current_voice_properties[ACSS.GAIN] = new_volume
        msg = f"{self._LOG_PREFIX}: Volume set to {new_volume}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _change_default_speech_pitch_range(self, step: float, decrease: bool = False) -> None:
        delta = step * (-1 if decrease else 1)
        pitch_range = self._default_voice.get(ACSS.PITCH_RANGE, 5)
        new_pitch_range = max(0, min(10, pitch_range + delta))
        self._default_voice[ACSS.PITCH_RANGE] = new_pitch_range
        self._current_voice_properties[ACSS.PITCH_RANGE] = new_pitch_range
        msg = f"{self._LOG_PREFIX}: Pitch range set to {new_pitch_range}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def increase_speech_rate(self, step: int = 5) -> None:
        """Increases the speech rate."""

        self._change_default_speech_rate(step)

    def decrease_speech_rate(self, step: int = 5) -> None:
        """Decreases the speech rate."""

        self._change_default_speech_rate(step, decrease=True)

    def increase_speech_pitch(self, step: float = 0.5) -> None:
        """Increases the speech pitch."""

        self._change_default_speech_pitch(step)

    def decrease_speech_pitch(self, step: float = 0.5) -> None:
        """Decreases the speech pitch."""

        self._change_default_speech_pitch(step, decrease=True)

    def increase_speech_volume(self, step: float = 0.5) -> None:
        """Increases the speech volume."""

        self._change_default_speech_volume(step)

    def decrease_speech_volume(self, step: float = 0.5) -> None:
        """Decreases the speech volume."""

        self._change_default_speech_volume(step, decrease=True)

    def increase_speech_inflection(self, step: float = 0.5) -> None:
        """Increases the speech inflection (pitch range)."""

        self._change_default_speech_pitch_range(step)

    def decrease_speech_inflection(self, step: float = 0.5) -> None:
        """Decreases the speech inflection (pitch range)."""

        self._change_default_speech_pitch_range(step, decrease=True)

    def update_capitalization_style(self, style: str) -> None:
        """Updates the capitalization style used by the speech server."""

    def update_punctuation_level(self, level: PunctuationStyle) -> None:
        """Punctuation level changed, inform this speechServer."""

        self._current_punctuation_level = level

    def stop(self) -> None:
        """Stops ongoing speech and flushes the queue."""

    def shutdown(self) -> None:
        """Shuts down the speech engine."""

    def reset(self) -> None:
        """Resets the speech engine."""

    def is_responsive(self, timeout: float = 2.0) -> bool:
        """Returns True if the server answers a probe within timeout seconds."""

        return True

    def clear_cached_voice_properties(self) -> None:
        """Clear cached voice properties to force reapplication on next speech."""

        msg = f"{self._LOG_PREFIX}: Clearing cached voice properties"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._current_voice_properties.clear()

    def get_output_module(self) -> str:
        """Returns the output module associated with this speech server."""

        return ""

    def set_output_module(self, module_id: str) -> None:
        """Sets the output module associated with this speech server."""

    def list_output_modules(self) -> tuple[str, ...]:
        """Return names of available output modules as a tuple of strings."""

        return ()

    def get_voice_families_for_language(
        self,
        language: str,
        dialect: str = "",
        variant: str | None = None,
        maximum: int | None = None,
    ) -> list[tuple[str, str, str | None]]:
        """Returns the families for language available in the current synthesizer."""

        return []

    def _filter_voices_for_language(
        self,
        voices: tuple[tuple[str, str, str | None], ...] | list[tuple[str, str, str | None]],
        target_language: str,
        target_dialect: str,
        variant: str | None = None,
        maximum: int | None = None,
    ) -> tuple[list[tuple[str, str, str | None]], list[tuple[str, str, str | None]]]:
        """Filters voices by language and dialect, returning candidates and fallbacks."""

        candidates: list[tuple[str, str, str | None]] = []
        fallbacks: list[tuple[str, str, str | None]] = []
        for voice in voices:
            normalized_language, normalized_dialect = self._normalized_language_and_dialect(
                voice[1],
            )
            if normalized_language != target_language:
                continue
            if variant not in (None, "none", "None") and voice[2] != variant:
                continue
            if normalized_dialect == target_dialect or (
                not normalized_dialect and target_dialect == normalized_language
            ):
                candidates.append(voice)
            elif not target_dialect:
                if normalized_dialect == target_language:
                    candidates.append(voice)
                if len(normalized_dialect) == 2:
                    fallbacks.append(voice)
            if maximum is not None and len(candidates) >= maximum:
                break
        return candidates, fallbacks

    def _get_language_and_dialect(self, acss_family: dict[str, Any] | None) -> tuple[str, str]:
        """Returns the language and dialect from the ACSS family dictionary."""

        if acss_family is None:
            acss_family = {}

        language = acss_family.get(VoiceFamily.LANG)
        dialect = acss_family.get(VoiceFamily.DIALECT)

        if not language:
            language, dialect = language_utilities.get_current_language_and_dialect()

        return str(language), str(dialect)

    def _normalized_language_and_dialect(self, language: str, dialect: str = "") -> tuple[str, str]:
        """Attempts to ensure consistency across inconsistent formats."""

        if "-" in language:
            normalized_language = language.split("-", 1)[0].lower()
            normalized_dialect = language.split("-", 1)[-1].lower()
        else:
            normalized_language = language.lower()
            normalized_dialect = dialect.lower()

        return normalized_language, normalized_dialect

    def get_language_and_dialect(self, acss_family: dict[str, Any] | None) -> tuple[str, str]:
        """Returns the language and dialect from the ACSS family dictionary."""

        return self._get_language_and_dialect(acss_family)



# Orca
#
# Copyright 2023 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Module for performing accessible actions via a list"""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, GLib, Gtk

from . import (
    action_presenter_command_definitions,
    dbus_service,
    debug,
    focus_manager,
    guilabels,
    input_event,
    messages,
    presentation_manager,
    script_manager,
)
from .ax_action import AXAction
from .ax_object import AXObject
from .ax_utilities import AXUtilities
from .extension import Extension

if TYPE_CHECKING:
    from collections.abc import Callable

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .command import Command
    from .scripts import default


class ActionPresenter(Extension):
    """Provides a list for performing accessible actions on an object."""

    GROUP_LABEL = guilabels.KB_GROUP_ACTIONS

    def __init__(self) -> None:
        self._gui: ActionList | None = None
        self._obj: Atspi.Accessible | None = None
        self._window: Atspi.Accessible | None = None
        super().__init__()

    def _get_commands(self) -> list[Command]:
        return action_presenter_command_definitions.get_commands(self)

    def _restore_focus(self) -> None:
        """Restores focus to the object associated with the actions list."""

        tokens = ["ACTION PRESENTER: Restoring focus to", self._obj, "in", self._window]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        reason = "Action Presenter list is being destroyed"
        app = AXUtilities.get_application(self._obj)
        script = script_manager.get_manager().get_script(app, self._obj)
        script_manager.get_manager().set_active_script(script, reason)

        manager = focus_manager.get_manager()
        manager.clear_state(reason)
        manager.set_active_window(self._window)
        manager.set_locus_of_focus(None, self._obj)

    def _clear_gui_and_restore_focus(self) -> None:
        """Clears the GUI reference and then restores focus."""

        self._gui = None
        self._restore_focus()

    def _perform_action(self, action: str) -> None:
        """Attempts to perform the named action."""

        if self._gui is None:
            msg = "ACTION PRESENTER: _perform_action called when self._gui is None."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        self._gui.hide()
        result = AXUtilities.do_named_action(self._obj, action)
        tokens = ["ACTION PRESENTER: Performing", action, "on", self._obj, "succeeded:", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        GLib.idle_add(self._gui.destroy)

    @dbus_service.command
    def show_actions_list(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Shows a list of all the accessible actions exposed by the focused object."""

        tokens = [
            "ACTION PRESENTER: show_actions_list. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        manager = focus_manager.get_manager()
        obj = manager.get_active_mode_and_object_of_interest()[1] or manager.get_locus_of_focus()
        if obj is None:
            full = messages.LOCATION_NOT_FOUND_FULL
            brief = messages.LOCATION_NOT_FOUND_BRIEF
            presentation_manager.get_manager().present_message(full, brief)
            return True

        actions = {}
        n_actions = AXAction.get_n_actions(obj)
        for i in range(n_actions):
            name = AXAction.get_action_name(obj, i, n_actions)
            localized_name = AXAction.get_action_localized_name(obj, i, n_actions)
            description = AXAction.get_action_description(obj, i, n_actions)
            tokens = [
                f"ACTION PRESENTER: Action {i} on",
                obj,
                f": '{name}' localized name: '{localized_name}' ",
                f"localized description: '{description}'",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            actions[name] = localized_name or description or name

        if not actions.items():
            name = AXObject.get_name(obj) or AXUtilities.get_localized_role_name(obj)
            presentation_manager.get_manager().present_message(messages.NO_ACTIONS_FOUND_ON % name)
            return True

        self._obj = obj
        self._window = manager.get_active_window()
        self._gui = ActionList(actions, self._perform_action, self._clear_gui_and_restore_focus)
        self._gui.show_gui()
        return True


class ActionList(Gtk.Window):
    """A Gtk.Window containing a Gtk.ListBox of accessible actions."""

    def __init__(
        self,
        actions: dict[str, str],
        action_handler: Callable[[str], None],
        cleanup_handler: Callable[[], None],
    ) -> None:
        super().__init__(window_position=Gtk.WindowPosition.MOUSE, transient_for=None)
        self.set_title(guilabels.ACTIONS_LIST)
        self.set_decorated(False)

        self.connect("destroy", self._on_hidden)
        self.on_option_selected = action_handler
        self.on_list_hidden = cleanup_handler

        self._list_box = Gtk.ListBox()
        self._list_box.set_selection_mode(Gtk.SelectionMode.SINGLE)
        self._list_box.connect("row-activated", self._on_row_activated)
        self._list_box.set_margin_top(5)
        self._list_box.set_margin_bottom(5)

        for name, description in actions.items():
            row = Gtk.ListBoxRow()
            label = Gtk.Label(label=description, xalign=0)
            label.set_margin_start(10)
            label.set_margin_end(10)
            row.add(label)  # pylint: disable=no-member
            row._action_name = name
            self._list_box.add(row)  # pylint: disable=no-member

        self.add(self._list_box)  # pylint: disable=no-member

        self.connect("key-press-event", self._on_key_press)

    def _on_key_press(self, _widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
        """Handles key presses for the window, e.g. Escape to close."""

        if event.keyval == Gdk.KEY_Escape:
            self.destroy()
            return True
        return False

    def _on_row_activated(self, _list_box: Gtk.ListBox, row: Gtk.ListBoxRow) -> None:
        """Handler for the 'row-activated' signal of the Gtk.ListBox"""

        action_name = getattr(row, "_action_name", None)
        if action_name:
            self.on_option_selected(action_name)

    def _on_hidden(self, *_args: tuple[Any, ...]) -> None:
        """Handler for the 'destroy' window signal"""

        msg = "ACTION PRESENTER: ActionList destroyed"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self.on_list_hidden()

    def show_gui(self) -> None:
        """Shows the window"""

        self.show_all()  # pylint: disable=no-member
        self.present_with_time(time.time())
        self._list_box.grab_focus()


_presenter = ActionPresenter()


def get_presenter() -> ActionPresenter:
    """Returns the action presenter."""

    return _presenter



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grids for sound presentation support."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from . import (
    gsettings_registry,
    guilabels,
    orca_gui_helpers,
    preferences_grid_base,
    sound_presenter,
)

if TYPE_CHECKING:
    import gi

    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk

    from .sound_presenter import SoundPresenter


class SoundProgressBarsPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Sound Progress Bars preferences page."""

    _documentation_summary = (
        "Use these settings to decide whether progress-bar changes are indicated with beeps."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for sound progress-bar preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.PROGRESS_BARS,
            panel_id="sound_presenter.progress-bars",
            description=(
                "Progress bar settings control whether Orca indicates progress-bar "
                "changes with beeps, how often beeps occur, and which progress bars apply."
            ),
            schema="sound",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.GENERAL_BEEP_UPDATES,
                    kind="switch",
                    summary="Controls whether Orca periodically beeps for progress-bar updates.",
                    schema="sound",
                    key=sound_presenter.SoundPresenter.KEY_BEEP_PROGRESS_BAR_UPDATES,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.GENERAL_FREQUENCY_SECS,
                    kind="spin",
                    summary="Sets how often Orca beeps for progress-bar updates.",
                    schema="sound",
                    key=sound_presenter.SoundPresenter.KEY_PROGRESS_BAR_BEEP_INTERVAL,
                    minimum="0",
                    maximum="100",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.GENERAL_APPLIES_TO,
                    kind="combo",
                    summary="Selects which progress bars produce beep updates.",
                    schema="sound",
                    key=sound_presenter.SoundPresenter.KEY_PROGRESS_BAR_BEEP_VERBOSITY,
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PROGRESS_BAR_ALL,
                            summary="Progress bars in all applications.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PROGRESS_BAR_APPLICATION,
                            summary="Progress bars in the active application.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PROGRESS_BAR_WINDOW,
                            summary="Progress bars in the active window.",
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, presenter: SoundPresenter) -> None:
        controls: list[preferences_grid_base.ControlType] = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.GENERAL_BEEP_UPDATES,
                getter=presenter.get_beep_progress_bar_updates,
                setter=presenter.set_beep_progress_bar_updates,
                prefs_key=sound_presenter.SoundPresenter.KEY_BEEP_PROGRESS_BAR_UPDATES,
            ),
            preferences_grid_base.IntRangePreferenceControl(
                label=guilabels.GENERAL_FREQUENCY_SECS,
                getter=presenter.get_progress_bar_beep_interval,
                setter=presenter.set_progress_bar_beep_interval,
                prefs_key=sound_presenter.SoundPresenter.KEY_PROGRESS_BAR_BEEP_INTERVAL,
                minimum=0,
                maximum=100,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.GENERAL_APPLIES_TO,
                getter=presenter.get_progress_bar_beep_verbosity,
                setter=presenter.set_progress_bar_beep_verbosity,
                prefs_key=sound_presenter.SoundPresenter.KEY_PROGRESS_BAR_BEEP_VERBOSITY,
                options=[
                    guilabels.PROGRESS_BAR_ALL,
                    guilabels.PROGRESS_BAR_APPLICATION,
                    guilabels.PROGRESS_BAR_WINDOW,
                ],
                values=[
                    sound_presenter.ProgressBarVerbosity.ALL.value,
                    sound_presenter.ProgressBarVerbosity.APPLICATION.value,
                    sound_presenter.ProgressBarVerbosity.WINDOW.value,
                ],
            ),
        ]

        super().__init__(guilabels.PROGRESS_BARS, controls)


class SoundPreferencesGrid(preferences_grid_base.PreferencesGridBase):
    """GtkGrid containing the Sound preferences page with nested stack navigation."""

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for sound output preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.SOUND,
            panel_id="manual.sound-output",
            description=(
                "Sound settings control Orca's non-speech audio output. The list contains "
                "sound-related settings and pages.\n\n"
                "Activate a row, or press Right Arrow, to open its settings. Press "
                "Left Arrow, Escape, or Alt+Left to return to the list."
            ),
            show_available_settings=False,
            schema="sound",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SOUND_ENABLE_SOUND_SUPPORT,
                    kind="switch",
                    summary="Turns sound output on or off.",
                    schema="sound",
                    key=sound_presenter.SoundPresenter.KEY_ENABLED,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SOUND_VOLUME,
                    kind="number",
                    summary="Sets the volume for Orca sounds.",
                    schema="sound",
                    key=sound_presenter.SoundPresenter.KEY_VOLUME,
                    minimum="0.0",
                    maximum="1.0",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.PROGRESS_BARS,
                    kind="page",
                    summary="Opens settings for when Orca beeps for progress-bar updates.",
                ),
            ),
        )

    def __init__(
        self,
        presenter: SoundPresenter,
        title_change_callback: preferences_grid_base.Callable[[str], None] | None = None,
    ) -> None:
        super().__init__(guilabels.SOUND)
        self._presenter = presenter
        self._initializing = True
        self._title_change_callback = title_change_callback

        self._progress_bars_grid = SoundProgressBarsPreferencesGrid(presenter)
        self._volume_scale: Gtk.Scale | None = None
        self._volume_listbox: preferences_grid_base.PreferencesFocusManagedListBox | None = None

        self._build()
        self._initializing = False

    def _build(self) -> None:
        """Build the nested stack UI."""

        row = 0

        categories = [
            (guilabels.PROGRESS_BARS, "progress-bars", self._progress_bars_grid),
        ]

        enable_listbox, stack, _categories_listbox = self._create_child_grid_preferences_stack(
            enable_label=guilabels.SOUND_ENABLE_SOUND_SUPPORT,
            enable_getter=self._presenter.get_sound_is_enabled,
            enable_setter=self._presenter.set_sound_is_enabled,
            categories=categories,
            title_change_callback=self._title_change_callback,
            main_title=guilabels.SOUND,
        )

        self.attach(enable_listbox, 0, row, 1, 1)
        row += 1

        # Volume slider on the main page
        self._volume_listbox = preferences_grid_base.PreferencesFocusManagedListBox()

        volume_row, self._volume_scale, _volume_label = orca_gui_helpers.create_range_slider_row(
            guilabels.SOUND_VOLUME,
            self._presenter.get_sound_volume(),
            0.0,
            1.0,
            changed_handler=self._on_volume_changed,
            include_top_separator=False,
            digits=1,
            steps=10,
            page_steps=10,
        )
        self._volume_listbox.add_row_with_widget(volume_row, self._volume_scale)
        self._volume_listbox.set_sensitive(self._presenter.get_sound_is_enabled())

        self.attach(self._volume_listbox, 0, row, 1, 1)
        row += 1

        self.attach(stack, 0, row, 1, 1)

    def _on_volume_changed(self, scale: Gtk.Scale) -> None:
        """Handle volume slider change."""

        if self._initializing:
            return
        value = scale.get_value()
        self._presenter.set_sound_volume(value)
        self._has_unsaved_changes = True

    def _on_multipage_enable_toggled(
        self,
        switch: Gtk.Switch,
        setter: preferences_grid_base.Callable[[bool], preferences_grid_base.Any],
    ) -> None:
        """Handle enable switch toggle - also controls volume slider sensitivity."""

        super()._on_multipage_enable_toggled(switch, setter)
        if self._volume_listbox is not None:
            self._volume_listbox.set_sensitive(switch.get_active())

    def _on_multipage_category_activated(self, listbox: Gtk.ListBox, row: Gtk.ListBoxRow) -> None:
        """Handle category activation - also hide volume slider."""

        super()._on_multipage_category_activated(listbox, row)
        if self._volume_listbox is not None:
            self._volume_listbox.hide()

    def multipage_show_categories(self) -> None:
        """Switch back to categories view - also show volume slider."""

        super().multipage_show_categories()
        if self._volume_listbox is not None:
            self._volume_listbox.show()

    def on_becoming_visible(self) -> None:
        """Reset to the categories view when this grid becomes visible."""

        self.multipage_on_becoming_visible()

    def reload(self) -> None:
        """Fetch fresh values and update UI."""

        self._initializing = True
        self._has_unsaved_changes = False

        enabled = self._presenter.get_sound_is_enabled()
        if self._volume_listbox is not None:
            self._volume_listbox.set_sensitive(enabled)
        if self._volume_scale is not None:
            self._volume_scale.set_value(self._presenter.get_sound_volume())
        self._progress_bars_grid.reload()

        self._initializing = False

    def save_settings(self, profile: str = "", app_name: str = "") -> dict:
        """Persist staged values."""

        result: dict[str, Any] = {}
        result["enabled"] = self._presenter.get_sound_is_enabled()
        result["volume"] = self._presenter.get_sound_volume()
        result.update(self._progress_bars_grid.save_settings())

        if profile:
            skip = not app_name and profile == "default"
            gsettings_registry.get_registry().save_schema("sound", result, profile, app_name, skip)

        return result

    def refresh(self) -> None:
        """Update widgets from staged values."""

        self._initializing = True
        if self._volume_scale is not None:
            self._volume_scale.set_value(self._presenter.get_sound_volume())
        self._progress_bars_grid.refresh()
        self._initializing = False

    def has_changes(self) -> bool:
        """Return True if any child grid has unsaved changes."""

        return self._progress_bars_grid.has_changes() or self._has_unsaved_changes



# Orca
#
# Copyright 2004-2008 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.


"""Custom script for The notification daemon."""

from __future__ import annotations

from typing import TYPE_CHECKING

from orca import messages, notification_presenter, presentation_manager
from orca.ax_text import AXText
from orca.ax_utilities import AXUtilities
from orca.scripts import default

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class Script(default.Script):
    """Custom script for The notification daemon."""

    def _on_window_created(self, event: Atspi.Event) -> bool:
        """Callback for window:create accessibility events."""

        texts = [AXText.get_all_text(acc) for acc in AXUtilities.find_all_labels(event.source)]
        text = f"{messages.NOTIFICATION} {' '.join(texts)}"

        presenter = presentation_manager.get_manager()
        presenter.speak_accessible_text(event.source, text)
        presenter.present_braille_message(text)
        notification_presenter.get_presenter().save_notification(text)



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for typing echo support."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import guilabels, preferences_grid_base, typing_echo_presenter

if TYPE_CHECKING:
    from .typing_echo_presenter import TypingEchoPresenter


class TypingEchoPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Typing Echo preferences page."""

    _gsettings_schema = "typing-echo"
    _documentation_summary = (
        "Use these settings to choose which typed keys, characters, words, and sentences "
        "Orca repeats."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for echo preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.ECHO,
            panel_id="typing_echo_presenter.echo",
            description=guilabels.ECHO_INFO,
            schema="typing-echo",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.ECHO_ENABLE_KEY_ECHO,
                    kind="switch",
                    summary=(
                        "Controls whether Orca speaks keys as they are pressed. When key "
                        "echo is enabled, use Keys to Echo to choose which key types Orca "
                        "speaks."
                    ),
                    schema="typing-echo",
                    key=typing_echo_presenter.TypingEchoPresenter.KEY_KEY_ECHO,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.ECHO_KEYS_TO_ECHO,
                    kind="group",
                    summary="Controls which types of keys Orca speaks when they are pressed.",
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_ALPHABETIC_KEYS,
                            kind="switch",
                            summary="Controls whether Orca speaks keys such as A, B, and C.",
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_ALPHABETIC_KEYS,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_NUMERIC_KEYS,
                            kind="switch",
                            summary="Controls whether Orca speaks keys such as 1, 2, and 3.",
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_NUMERIC_KEYS,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_PUNCTUATION_KEYS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks keys such as percent, "
                                "semicolon, and question mark."
                            ),
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_PUNCTUATION_KEYS,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_SPACE,
                            kind="switch",
                            summary="Controls whether Orca speaks Space when it is pressed.",
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_SPACE,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_MODIFIER_KEYS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks modifier keys such as Shift, "
                                "Control, Alt, and Meta."
                            ),
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_MODIFIER_KEYS,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_FUNCTION_KEYS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks function keys from F1 through F12."
                            ),
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_FUNCTION_KEYS,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_ACTION_KEYS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks action keys such as Backspace, "
                                "Delete, Return, Escape, Tab, Page Up, and Page Down."
                            ),
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_ACTION_KEYS,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_NAVIGATION_KEYS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks navigation keys such as Left, "
                                "Right, Up, Down, Home, and End."
                            ),
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_NAVIGATION_KEYS,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_DIACRITICAL_KEYS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks diacritical keys, such as "
                                "keys used to add accents to letters."
                            ),
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_DIACRITICAL_KEYS,
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.ECHO_TYPING_ECHO,
                    kind="group",
                    summary=(
                        "Controls whether Orca speaks text inserted into the document as you type."
                    ),
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_CHARACTER,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks each inserted character. This "
                                "can include accented letters and other symbols that do not "
                                "have a dedicated key."
                            ),
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_CHARACTER_ECHO,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_WORD,
                            kind="switch",
                            summary="Controls whether Orca speaks each completed word.",
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_WORD_ECHO,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ECHO_SENTENCE,
                            kind="switch",
                            summary=("Controls whether Orca speaks each completed sentence."),
                            schema="typing-echo",
                            key=typing_echo_presenter.TypingEchoPresenter.KEY_SENTENCE_ECHO,
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, presenter: TypingEchoPresenter) -> None:
        self._enable_key_echo_control = preferences_grid_base.BooleanPreferenceControl(
            label=guilabels.ECHO_ENABLE_KEY_ECHO,
            getter=presenter.get_key_echo_enabled,
            setter=presenter.set_key_echo_enabled,
            prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_KEY_ECHO,
        )

        controls = [
            self._enable_key_echo_control,
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_ALPHABETIC_KEYS,
                getter=presenter.get_alphabetic_keys_enabled,
                setter=presenter.set_alphabetic_keys_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_ALPHABETIC_KEYS,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_NUMERIC_KEYS,
                getter=presenter.get_numeric_keys_enabled,
                setter=presenter.set_numeric_keys_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_NUMERIC_KEYS,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_PUNCTUATION_KEYS,
                getter=presenter.get_punctuation_keys_enabled,
                setter=presenter.set_punctuation_keys_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_PUNCTUATION_KEYS,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_SPACE,
                getter=presenter.get_space_enabled,
                setter=presenter.set_space_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_SPACE,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_MODIFIER_KEYS,
                getter=presenter.get_modifier_keys_enabled,
                setter=presenter.set_modifier_keys_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_MODIFIER_KEYS,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_FUNCTION_KEYS,
                getter=presenter.get_function_keys_enabled,
                setter=presenter.set_function_keys_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_FUNCTION_KEYS,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_ACTION_KEYS,
                getter=presenter.get_action_keys_enabled,
                setter=presenter.set_action_keys_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_ACTION_KEYS,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_NAVIGATION_KEYS,
                getter=presenter.get_navigation_keys_enabled,
                setter=presenter.set_navigation_keys_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_NAVIGATION_KEYS,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_DIACRITICAL_KEYS,
                getter=presenter.get_diacritical_keys_enabled,
                setter=presenter.set_diacritical_keys_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_DIACRITICAL_KEYS,
                member_of=guilabels.ECHO_KEYS_TO_ECHO,
                determine_sensitivity=presenter.get_key_echo_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_CHARACTER,
                getter=presenter.get_character_echo_enabled,
                setter=presenter.set_character_echo_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_CHARACTER_ECHO,
                member_of=guilabels.ECHO_TYPING_ECHO,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_WORD,
                getter=presenter.get_word_echo_enabled,
                setter=presenter.set_word_echo_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_WORD_ECHO,
                member_of=guilabels.ECHO_TYPING_ECHO,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ECHO_SENTENCE,
                getter=presenter.get_sentence_echo_enabled,
                setter=presenter.set_sentence_echo_enabled,
                prefs_key=typing_echo_presenter.TypingEchoPresenter.KEY_SENTENCE_ECHO,
                member_of=guilabels.ECHO_TYPING_ECHO,
            ),
        ]

        self._presenter = presenter
        super().__init__(guilabels.ECHO, controls, info_message=guilabels.ECHO_INFO)
